"""Tests for GeoInterpolator."""

import unittest

import numpy as np
import pytest
from pyresample.geometry import SwathDefinition

from geotiepoints.geointerpolator import GeoInterpolator, GeoGridInterpolator, GeoSplineInterpolator

TIES_EXP1 = np.array([[6384905.78040055, 6381081.08333225, 6371519.34066148,
                       6328950.00792935, 6253610.69157758, 6145946.19489936,
                       6124413.29556372],
                      [6377591.95940176, 6370997., 6354509.6014956,
                       6305151.62592155, 6223234.99818839, 6109277.14889072,
                       6086485.57903118],
                      [6359307.40690478, 6345786.79166939, 6311985.2535809,
                       6245655.67090206, 6147295.76471541, 6017604.5338691,
                       5991666.28769983],
                      [6351993.58590599, 6335702.70833714, 6294975.51441502,
                       6221857.28889426, 6116920.07132621, 5980935.48786045,
                       5953738.5711673],
                      [6338032.26190294, 6320348.4990906, 6276139.09205974,
                       6199670.56624433, 6091551.90273768, 5952590.38414781,
                       5924798.08042984],
                      [6290665.5946295, 6270385.16249031, 6219684.08214232,
                       6137100.75832981, 6023313.2794414, 5879194.72399075,
                       5850371.01290062],
                      [6172248.92644589, 6145476.82098957, 6078546.55734877,
                       5980676.23854351, 5852716.72120069, 5695705.57359808,
                       5664303.34407756],
                      [6124882.25917245, 6095513.48438928, 6022091.54743135,
                       5918106.430629, 5784478.09790441, 5622309.91344102,
                       5589876.27654834]])

TIES_EXP2 = np.array([[6372937.31273379, 6370997., 6366146.21816553,
                       6351605.98629588, 6327412.61244969, 6293626.50067273,
                       6286869.27831734],
                      [6353136.46335726, 6345786.79166939, 6327412.61244969,
                       6299445.69529922, 6261968.60390423, 6215087.60607344,
                       6205711.40650728]])

TIES_EXP4 = np.array([[6381081.08333225, 6381639.66045187, 6372470.10269454,
                       6353590.21586788, 6325042.05851245],
                      [6370997., 6366146.21816553, 6351605.98629588,
                       6327412.61244969, 6293626.50067273],
                      [6345786.79166939, 6327412.61244969, 6299445.69529922,
                       6261968.60390423, 6215087.60607344],
                      [6335702.70833714, 6311919.17016336, 6278581.57890056,
                       6235791.00048604, 6183672.04823372]])

TIES_EXP5 = np.array([[6381081.08333225, 6371519.34066148, 6328950.00792935,
                       6253610.69157758, 6145946.19489936],
                      [6370997., 6354509.6014956, 6305151.62592155,
                       6223234.99818839, 6109277.14889072],
                      [6345786.79166939, 6311985.2535809, 6245655.67090206,
                       6147295.76471541, 6017604.5338691],
                      [6270385.16249031, 6219684.08214232, 6137100.75832981,
                       6023313.2794414, 5879194.72399075],
                      [6145476.82098957, 6078546.55734877, 5980676.23854351,
                       5852716.72120069, 5695705.57359808],
                      [6095513.48438928, 6022091.54743135, 5918106.430629,
                       5784478.09790441, 5622309.91344102]])

TIES_EXP6 = np.array([[6381081.08333225, 6371519.34066148, 6328950.00792935,
                       6253610.69157758, 6145946.19489936],
                      [6370997., 6354509.6014956, 6305151.62592155,
                       6223234.99818839, 6109277.14889072],
                      [6345786.79166939, 6311985.2535809, 6245655.67090206,
                       6147295.76471541, 6017604.5338691],
                      [6335702.70833714, 6294975.51441502, 6221857.28889426,
                       6116920.07132621, 5980935.48786045],
                      [6320348.4990906, 6276139.09205974, 6199670.56624433,
                       6091551.90273768, 5952590.38414781],
                      [6270385.16249031, 6219684.08214232, 6137100.75832981,
                       6023313.2794414, 5879194.72399075],
                      [6145476.82098957, 6078546.55734877, 5980676.23854351,
                       5852716.72120069, 5695705.57359808],
                      [6095513.48438928, 6022091.54743135, 5918106.430629,
                       5784478.09790441, 5622309.91344102]])

TIES_EXP7 = np.array([[6372937.31273379, 6370997., 6366146.21816553,
                       6351605.98629588, 6327412.61244969, 6293626.50067273,
                       6286869.27831734],
                      [6353136.46335726, 6345786.79166939, 6327412.61244969,
                       6299445.69529922, 6261968.60390423, 6215087.60607344,
                       6205711.40650728]])


class TestGeoInterpolator(unittest.TestCase):
    """Class for unit testing the ancillary interpolation functions."""

    def test_fillborders(self):
        """Test filling borders."""
        lons = np.arange(20).reshape((4, 5), order="F")
        lats = np.arange(20).reshape((4, 5), order="C")
        lines = np.array([2, 7, 12, 17]) / 5.0
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(20) / 5.0
        hcols = np.arange(24)
        satint = GeoInterpolator(
            (lons, lats), (lines, cols), (hlines, hcols), chunk_size=10)
        satint.fill_borders('x', 'y')

        np.testing.assert_allclose(satint.tie_data[0], TIES_EXP1)
        np.testing.assert_allclose(satint.row_indices, np.array(
            [0, 2, 7, 9, 10, 12, 17, 19]) / 5.0)
        self.assertTrue(
            np.allclose(satint.col_indices, np.array([0, 2, 7, 12, 17, 22, 23])))

    def test_extrapolate_cols(self):
        """Test extrapolating columns."""
        lons = np.arange(10).reshape((2, 5), order="F")
        lats = np.arange(10).reshape((2, 5), order="C")
        lines = np.array([2, 7])
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(10)
        hcols = np.arange(24)
        satint = GeoInterpolator((lons, lats), (lines, cols), (hlines, hcols))

        self.assertTrue(np.allclose(satint._extrapolate_cols(satint.tie_data[0]),
                                    TIES_EXP2))

    def test_fill_col_borders(self):
        """Test filling the column borders."""
        lons = np.arange(10).reshape((2, 5), order="F")
        lats = np.arange(10).reshape((2, 5), order="C")
        lines = np.array([2, 7])
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(10)
        hcols = np.arange(24)
        satint = GeoInterpolator((lons, lats), (lines, cols), (hlines, hcols))
        satint._fill_col_borders()
        np.testing.assert_allclose(satint.tie_data[0], TIES_EXP7)
        np.testing.assert_allclose(satint.col_indices,
                                   np.array([0, 2, 7, 12, 17, 22, 23]))

    def test_extrapolate_rows(self):
        """Test extrapolation of rows."""
        lons = np.arange(10).reshape((2, 5), order="F")
        lats = np.arange(10).reshape((2, 5), order="C")
        lines = np.array([2, 7])
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(10)
        hcols = np.arange(24)
        satint = GeoInterpolator((lons, lats), (lines, cols), (hlines, hcols))
        np.testing.assert_allclose(satint._extrapolate_rows(satint.tie_data[0],
                                                            hlines, -0.4, 9.4),
                                   TIES_EXP4)

    def test_fill_row_borders(self):
        """Test filling the row borders."""
        lons = np.arange(20).reshape((4, 5), order="F")
        lats = np.arange(20).reshape((4, 5), order="C")
        lines = np.array([2, 7, 12, 17]) / 5.0
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(20) / 5.0
        hcols = np.arange(24)
        satint = GeoInterpolator((lons, lats), (lines, cols), (hlines, hcols))
        satint._fill_row_borders()
        np.testing.assert_allclose(satint.tie_data[0],
                                   TIES_EXP5)
        np.testing.assert_allclose(satint.row_indices,
                                   np.array([0, 2, 7, 12, 17, 19]) / 5.0)
        satint = GeoInterpolator((lons, lats), (lines, cols),
                                 (hlines, hcols), chunk_size=10)
        satint._fill_row_borders()
        np.testing.assert_allclose(satint.tie_data[0],
                                   TIES_EXP6)
        np.testing.assert_allclose(satint.row_indices,
                                   np.array([0, 2, 7, 9, 10, 12, 17, 19]) / 5.0)


TIE_LONS = np.array([[1, 2, 3, 4],
                     [1, 2, 3, 4],
                     [1, 2, 3, 4],
                     [1, 2, 3, 4],
                     [1, 2, 3, 4]])

TIE_LATS = np.array([[1, 1, 1, 1],
                     [2, 2, 2, 2],
                     [3, 3, 3, 3],
                     [4, 4, 4, 4],
                     [5, 5, 5, 5]])


class TestGeoGridInterpolator:
    """Test the GeoGridInterpolator."""

    @pytest.mark.parametrize("args", ((TIE_LONS, TIE_LATS),
                                      [SwathDefinition(TIE_LONS, TIE_LATS)]
                                      ))
    def test_geogrid_interpolation(self, args):
        """Test that the interpolator works with both explicit tie-point arrays and swath definition objects."""
        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoGridInterpolator((y_points, x_points), *args)

        fine_x_points = np.arange(8)
        fine_y_points = np.arange(16)

        lons, lats = interpolator.interpolate((fine_y_points, fine_x_points))

        lons_expected = np.array([1., 2., 2.5, 3., 3.25, 3.5, 3.75, 4.])
        lats_expected = np.array([1., 2., 2.5, 3., 3.25, 3.5, 3.75, 4., 4.125,
                                  4.25, 4.375, 4.5, 4.625, 4.75, 4.875, 5.])

        np.testing.assert_allclose(lons[0, :], lons_expected, rtol=5e-5)
        np.testing.assert_allclose(lats[:, 0], lats_expected, rtol=5e-5)

    def test_geogrid_interpolation_counts_its_arguments(self):
        """Test that an arbitrary number of argument is not allowed in the interpolator."""
        with pytest.raises(ValueError):
            _ = GeoGridInterpolator((None, None), None, None, None)

    def test_geogrid_interpolation_to_shape(self):
        """Test that the interpolator works with both explicit tie-point arrays and swath definition objects."""
        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoGridInterpolator((y_points, x_points), TIE_LONS, TIE_LATS)

        lons, lats = interpolator.interpolate_to_shape((16, 8))

        lons_expected = np.array([1., 2., 2.5, 3., 3.25, 3.5, 3.75, 4.])
        lats_expected = np.array([1., 2., 2.5, 3., 3.25, 3.5, 3.75, 4., 4.125,
                                  4.25, 4.375, 4.5, 4.625, 4.75, 4.875, 5.])

        np.testing.assert_allclose(lons[0, :], lons_expected, rtol=5e-5)
        np.testing.assert_allclose(lats[:, 0], lats_expected, rtol=5e-5)

    def test_geogrid_interpolation_preserves_dtype(self):
        """Test that the interpolator works with both explicit tie-point arrays and swath definition objects."""
        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoGridInterpolator((y_points, x_points),
                                           TIE_LONS.astype(np.float32), TIE_LATS.astype(np.float32))

        lons, lats = interpolator.interpolate_to_shape((16, 8))

        assert lons.dtype == np.float32
        assert lats.dtype == np.float32

    def test_chunked_geogrid_interpolation(self):
        """Test that the interpolator works with both explicit tie-point arrays and swath definition objects."""
        dask = pytest.importorskip("dask")

        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoGridInterpolator((y_points, x_points),
                                           TIE_LONS.astype(np.float32), TIE_LATS.astype(np.float32))

        lons, lats = interpolator.interpolate_to_shape((16, 8), chunks=4)

        assert lons.chunks == ((4, 4, 4, 4), (4, 4))
        assert lats.chunks == ((4, 4, 4, 4), (4, 4))

        with dask.config.set({"array.chunk-size": 64}):

            lons, lats = interpolator.interpolate_to_shape((16, 8), chunks="auto")
            assert lons.chunks == ((4, 4, 4, 4), (4, 4))
            assert lats.chunks == ((4, 4, 4, 4), (4, 4))

    def test_geogrid_interpolation_can_extrapolate(self):
        """Test that the interpolator can also extrapolate given the right parameters."""
        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoGridInterpolator((y_points, x_points), TIE_LONS, TIE_LATS,
                                           bounds_error=False, fill_value=None)

        lons, lats = interpolator.interpolate_to_shape((16, 16), method="cubic")

        assert lons.shape == (16, 16)


class TestGeoSplineInterpolator:
    """Test the GeoGridInterpolator."""

    @pytest.mark.parametrize("args", ((TIE_LONS, TIE_LATS),
                                      [SwathDefinition(TIE_LONS, TIE_LATS)]
                                      ))
    def test_geospline_interpolation(self, args):
        """Test that the interpolator works with both explicit tie-point arrays and swath definition objects."""
        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoSplineInterpolator((y_points, x_points), *args, kx=1, ky=1)

        fine_x_points = np.arange(8)
        fine_y_points = np.arange(16)

        lons, lats = interpolator.interpolate((fine_y_points, fine_x_points))

        lons_expected = np.array([1., 2., 2.5, 3., 3.25, 3.5, 3.75, 4.])
        lats_expected = np.array([1., 2., 2.5, 3., 3.25, 3.5, 3.75, 4., 4.125,
                                  4.25, 4.375, 4.5, 4.625, 4.75, 4.875, 5.])

        np.testing.assert_allclose(lons[0, :], lons_expected, rtol=5e-5)
        np.testing.assert_allclose(lats[:, 0], lats_expected, rtol=5e-5)

    def test_geospline_interpolation_to_shape(self):
        """Test that the interpolator works with both explicit tie-point arrays and swath definition objects."""
        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoSplineInterpolator((y_points, x_points), TIE_LONS, TIE_LATS, kx=1, ky=1)

        lons, lats = interpolator.interpolate_to_shape((16, 8))

        lons_expected = np.array([1., 2., 2.5, 3., 3.25, 3.5, 3.75, 4.])
        lats_expected = np.array([1., 2., 2.5, 3., 3.25, 3.5, 3.75, 4., 4.125,
                                  4.25, 4.375, 4.5, 4.625, 4.75, 4.875, 5.])

        np.testing.assert_allclose(lons[0, :], lons_expected, rtol=5e-5)
        np.testing.assert_allclose(lats[:, 0], lats_expected, rtol=5e-5)

    def test_geospline_interpolation_preserves_dtype(self):
        """Test that the interpolator works with both explicit tie-point arrays and swath definition objects."""
        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoGridInterpolator((y_points, x_points),
                                           TIE_LONS.astype(np.float32), TIE_LATS.astype(np.float32))

        lons, lats = interpolator.interpolate_to_shape((16, 8))

        assert lons.dtype == np.float32
        assert lats.dtype == np.float32

    def test_chunked_geospline_interpolation(self):
        """Test that the interpolator works with both explicit tie-point arrays and swath definition objects."""
        dask = pytest.importorskip("dask")

        x_points = np.array([0, 1, 3, 7])
        y_points = np.array([0, 1, 3, 7, 15])

        interpolator = GeoGridInterpolator((y_points, x_points),
                                           TIE_LONS.astype(np.float32), TIE_LATS.astype(np.float32))

        lons, lats = interpolator.interpolate_to_shape((16, 8), chunks=4)

        assert lons.chunks == ((4, 4, 4, 4), (4, 4))
        assert lats.chunks == ((4, 4, 4, 4), (4, 4))

        with dask.config.set({"array.chunk-size": 64}):

            lons, lats = interpolator.interpolate_to_shape((16, 8), chunks="auto")
            assert lons.chunks == ((4, 4, 4, 4), (4, 4))
            assert lats.chunks == ((4, 4, 4, 4), (4, 4))



"""Tests for simple MODIS interpolators."""

import numpy as np
import pytest
import dask
import dask.array as da

from geotiepoints.simple_modis_interpolator import modis_1km_to_250m, modis_1km_to_500m
from .test_modisinterpolator import (
    assert_geodetic_distance,
    load_1km_lonlat_as_xarray_dask,
    load_1km_lonlat_as_dask,
    load_1km_lonlat_as_numpy,
    load_500m_lonlat_expected_as_xarray_dask,
    load_250m_lonlat_expected_as_xarray_dask,
)

from .utils import CustomScheduler


@pytest.mark.parametrize(
    ("input_func", "exp_func", "interp_func", "dist_max"),
    [
        (load_1km_lonlat_as_xarray_dask, load_500m_lonlat_expected_as_xarray_dask, modis_1km_to_500m, 16),
        (load_1km_lonlat_as_xarray_dask, load_250m_lonlat_expected_as_xarray_dask, modis_1km_to_250m, 27.35),
        (load_1km_lonlat_as_dask, load_500m_lonlat_expected_as_xarray_dask, modis_1km_to_500m, 16),
        (load_1km_lonlat_as_dask, load_250m_lonlat_expected_as_xarray_dask, modis_1km_to_250m, 27.35),
        (load_1km_lonlat_as_numpy, load_500m_lonlat_expected_as_xarray_dask, modis_1km_to_500m, 16),
        (load_1km_lonlat_as_numpy, load_250m_lonlat_expected_as_xarray_dask, modis_1km_to_250m, 27.35),
    ]
)
def test_basic_interp(input_func, exp_func, interp_func, dist_max):
    lon1, lat1 = input_func()
    lons_exp, lats_exp = exp_func()

    # when working with dask arrays, we shouldn't compute anything
    with dask.config.set(scheduler=CustomScheduler(0)):
        lons, lats = interp_func(lon1, lat1)

    if hasattr(lons, "compute"):
        lons, lats = da.compute(lons, lats)
    assert_geodetic_distance(lons, lats, lons_exp, lats_exp, dist_max)
    assert not np.any(np.isnan(lons))
    assert not np.any(np.isnan(lats))


def test_nonstandard_scan_size():
    lon1, lat1 = load_1km_lonlat_as_xarray_dask()
    # remove 1 row from the end
    lon1 = lon1[:-1]
    lat1 = lat1[:-1]

    pytest.raises(ValueError, modis_1km_to_250m, lon1, lat1)



"""Interpolate MODIS 1km navigation arrays to 250m and 500m resolutions.

The code used here is a rewrite of the IDL function ``MODIS_GEO_INTERP_250``
used by Liam Gumley. It has been modified to convert coordinates to cartesian
(X, Y, Z) coordinates first to avoid problems with the anti-meridian and poles.
This code was originally part of the CSPP Polar2Grid project, but has been
moved here for integration with Satpy and newer versions of Polar2Grid.

This algorithm differs from the one in ``modisinterpolator`` as it only
requires the original longitude and latitude arrays. This is useful in the
case of reading the 250m or 500m MODIS L1b files or any MODIS L2 files without
including the MOD03 geolocation file as there is no SensorZenith angle dataset
in these files.

"""

from ._modis_utils import scanline_mapblocks
from ._simple_modis_interpolator import interpolate_geolocation_cartesian as interp_cython


@scanline_mapblocks
def interpolate_geolocation_cartesian(lon_array, lat_array, coarse_resolution, fine_resolution):
    """Interpolate MODIS navigation from 1000m resolution to 250m.

    Python rewrite of the IDL function ``MODIS_GEO_INTERP_250`` but converts to cartesian (X, Y, Z) coordinates
    first to avoid problems with the anti-meridian/poles.

    Arguments:
        lon_array: Longitude data as a 2D numpy, dask, or xarray DataArray object.
            The input data is expected to represent 1000m geolocation.
        lat_array: Latitude data as a 2D numpy, dask, or xarray DataArray object.
            The input data is expected to represent 1000m geolocation.
        res_factor (int): Expansion factor for the function. Should be 2 for
            500m output or 4 for 250m output.

    Returns:
        A two-element tuple (lon, lat).

    """
    return interp_cython(
        lon_array, lat_array, coarse_resolution, fine_resolution
    )


def modis_1km_to_250m(lon1, lat1):
    """Interpolate MODIS geolocation from 1km to 250m resolution."""
    return interpolate_geolocation_cartesian(
        lon1,
        lat1,
        coarse_resolution=1000,
        fine_resolution=250,
    )


def modis_1km_to_500m(lon1, lat1):
    """Interpolate MODIS geolocation from 1km to 500m resolution."""
    return interpolate_geolocation_cartesian(
        lon1,
        lat1,
        coarse_resolution=1000,
        fine_resolution=500)



from __future__ import division

import numpy as np

from geotiepoints.multilinear_cython import multilinear_interpolation


def mlinspace(smin, smax, orders):
    if len(orders) == 1:
        res = np.atleast_2d(
            np.linspace(np.array(smin), np.array(smax), np.array(orders)))
        return res.copy()  # workaround for strange bug
    else:
        meshes = np.meshgrid(
            *[np.linspace(smin[i], smax[i], orders[i]) for i in range(len(orders))], indexing='ij')
        return np.vstack([l.flatten() for l in meshes])


class MultilinearInterpolator:
    """Multilinear interpolation.

    Methods
    -------
    smin, smax, orders : iterable objects
        Specifies the boundaries of a cartesian grid, with the number of points along each dimension.
    values : array_like (2D), optional
        Each line enumerates values taken by a function on the cartesian grid, with last index varying faster.


    Attributes
    ----------
    smin, smax, orders :
        Boundaries and number of points along each dimension.
    d : number of dimensions
    grid : array_like (2D)
        Enumerate the approximation grid. Each column contains coordinates of a point in R^d.
    values :
        Values on the grid to be interpolated.

    Example
    -------

    smin = [-1,-1]
    smax = [1,1]
    orders = [5,5]

    f = lambda x: np.vstack([np.sqrt( x[0,:]**2 + x[1,:]**2 ),
                             np.power( x[0,:]**3 + x[1,:]**3, 1.0/3.0 )])

    interp = MultilinearInterpolator(smin,smax,orders)
    interp.set_values( f(interp.grid) )

    random_points = np.random.random( (2, 1000) )

    interpolated_values = interp(random_points)
    exact_values = f(random_points)
    """

    __grid__ = None

    def __init__(self, smin, smax, orders, values=None, dtype=np.float64):
        self.smin = np.asarray(smin, dtype=dtype)
        self.smax = np.asarray(smax, dtype=dtype)
        self.orders = np.asarray(orders, dtype="long")
        self.d = len(orders)
        self.dtype = dtype
        if values is not None:
            self.set_values(values)

    @property
    def grid(self):
        if self.__grid__ is None:
            self.__grid__ = mlinspace(self.smin, self.smax, self.orders)
        return self.__grid__

    def set_values(self, values):
        self.values = np.ascontiguousarray(values, dtype=self.dtype)

    def interpolate(self, s):
        s = np.ascontiguousarray(s, dtype=self.dtype)
        a = multilinear_interpolation(
            self.smin, self.smax, self.orders, self.values, s)
        return a

    def __call__(self, s):
        return self.interpolate(s)



import numpy as np
# from memory_profiler import profile
from pandas import DataFrame, Series

class BasicSatelliteInterpolator(object):
    """
    Handles interpolation of geolocation data from a grid of tie points.  

    Currently, it is assumed, that tie-points reach out til the edges if the
    tiepoint grid. Methods for extrapolation need to be added if needed.

    Uses numpy and pandas.

    The constructor gets the tiepointed lat/lon data as *lat_data* and 
    *lon_data*.
    """

    def __init__(self, cols, rows, lat_data, lon_data):
        self.row_indices = rows
        self.col_indices = cols

        self.lon_tiepoint = lon_data
        self.lat_tiepoint = lat_data

        self.longitudes = None
        self.latitudes = None


    def _interp(self, data):
        """The interpolation method implemented here is a kind of a billinear
        interpolation. The input *data* field is first interpolated along the 
        rows and subsequently along its columns.

        The final size of the interpolated *data* field is determined by the 
        last indices in self.row_indices and self.col_indices.
        """
        row_interpol_data = self._interp_axis(data, 0)
        interpol_data = self._interp_axis(row_interpol_data, 1)

        return interpol_data


    # @profile
    def _interp_axis(self, data, axis):
        """The *data* field contains the data to be interpolated. It is
        expected that values reach out to the *data* boundaries.
        With *axis*=0 this method interpolates along rows and *axis*=1 it
        interpolates along colums.

        For column mode the *data* input is transposed before interpolation
        and subsequently transposed back.
        """
        if axis == 0:
            return self._pandas_interp(data, self.row_indices)
        
        if axis == 1:
            data_transposed = data.as_matrix().T
            data_interpol_transposed = self._pandas_interp(data_transposed, 
                                                            self.col_indices)
            data_interpol = data_interpol_transposed.as_matrix().T

            return data_interpol


    def _pandas_interp(self, data, indices):
        """The actual transformation based on the following stackoverflow 
        entry: http://stackoverflow.com/a/10465162
        """
        new_index = np.arange(indices[-1] + 1)

        data_frame = DataFrame(data, index=indices)
        data_frame_reindexed = data_frame.reindex(new_index)
        data_interpol = data_frame_reindexed.apply(Series.interpolate)

        del new_index
        del data_frame
        del data_frame_reindexed

        return data_interpol


    def interpolate(self):
        """Do the interpolation and return resulting longitudes and latitudes.
        """
        self.latitude = self._interp(self.lat_tiepoint)
        self.longitude = self._interp(self.lon_tiepoint)

        return self.latitude, self.longitude


"""Utilities for creating and checking tests."""


class CustomScheduler(object):
    """Scheduler raising an exception if data are computed too many times.

    This only makes sense to include when working with dask-based arrays. To
    use it::

        with dask.config.set(scheduler=CustomScheduler(2)):
            my_dask_arr.compute()  # allowed
            my_dask_arr.compute()  # 2nd call, not allowed, fails

    """

    def __init__(self, max_computes=1):
        """Set starting and maximum compute counts."""
        self.max_computes = max_computes
        self.total_computes = 0

    def __call__(self, dsk, keys, **kwargs):
        """Compute dask task and keep track of number of times we do so."""
        import dask
        self.total_computes += 1
        if self.total_computes > self.max_computes:
            raise RuntimeError("Too many dask computations were scheduled: "
                               "{}".format(self.total_computes))
        return dask.get(dsk, keys, **kwargs)



# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False, initializedcheck=False, nonecheck=False
cimport cython
import numpy as np
cimport numpy as np

np.import_array()

def test_func():
    cdef np.ndarray[float, ndim=2] arr = np.zeros((5, 5), dtype=np.float32)
    cdef float[:, ::1] arr_view = arr
    _run(arr_view)

cdef void _run(float[:, ::1] arr_view) noexcept nogil:
    cdef float[:, :] tmp = _get_upper_left_corner(arr_view)

cdef inline float[:, :] _get_upper_left_corner(float[:, ::1] arr) noexcept nogil:
    return arr[:1, :1]



"""Geographical interpolation (lon/lats)."""

import numpy as np
from geotiepoints.interpolator import Interpolator, MultipleGridInterpolator, MultipleSplineInterpolator


EARTH_RADIUS = 6370997.0


class GeoInterpolator(Interpolator):
    """Handles interpolation of geolocation from a grid of tie points.

    It is
    preferable to have tie-points out till the edges if the tiepoint grid, but
    a method is provided to extrapolate linearly the tiepoints to the borders
    of the grid. The extrapolation is done automatically if it seems necessary.

    Uses numpy, scipy, and optionally pyresample.

    The constructor takes in the tiepointed data as *data*, the
    *tiepoint_grid* and the desired *final_grid*. As optional arguments, one
    can provide *kx_* and *ky_* as interpolation orders (in x and y directions
    respectively), and the *chunksize* if the data has to be handled by pieces
    along the y axis (this affects how the extrapolator behaves). If
    *chunksize* is set, don't forget to adjust the interpolation orders
    accordingly: the interpolation is indeed done globaly (not chunkwise).

    """

    def __init__(self, lon_lat_data, *args, **kwargs):
        try:
            # Maybe it's a pyresample object ?
            self.lon_tiepoint = lon_lat_data.lons
            self.lat_tiepoint = lon_lat_data.lats
            xyz = lon_lat_data.get_cartesian_coords()
            tie_data = [xyz[:, :, 0], xyz[:, :, 1], xyz[:, :, 2]]
        except AttributeError:
            self.lon_tiepoint = lon_lat_data[0]
            self.lat_tiepoint = lon_lat_data[1]
            x__, y__, z__ = lonlat2xyz(self.lon_tiepoint, self.lat_tiepoint)
            tie_data = [x__, y__, z__]

        super().__init__(tie_data, *args, **kwargs)

    def interpolate(self):
        """Run the interpolation."""
        newx, newy, newz = super().interpolate()
        lon, lat = xyz2lonlat(newx, newy, newz)
        return lon, lat


def lonlat2xyz(lons, lats, radius=EARTH_RADIUS):
    """Convert lons and lats to cartesian coordinates."""
    lons_rad = np.deg2rad(lons)
    lats_rad = np.deg2rad(lats)
    x_coords = radius * np.cos(lats_rad) * np.cos(lons_rad)
    y_coords = radius * np.cos(lats_rad) * np.sin(lons_rad)
    z_coords = radius * np.sin(lats_rad)
    return x_coords, y_coords, z_coords


def xyz2lonlat(x__, y__, z__, radius=EARTH_RADIUS, thr=0.8, low_lat_z=True):
    """Get longitudes from cartesian coordinates."""
    lons = np.rad2deg(np.arccos(x__ / np.sqrt(x__ ** 2 + y__ ** 2))) * np.sign(y__)
    lats = np.sign(z__) * (90 - np.rad2deg(np.arcsin(np.sqrt(x__ ** 2 + y__ ** 2) / radius)))
    if low_lat_z:
        # if we are at low latitudes - small z, then get the
        # latitudes only from z. If we are at high latitudes (close to the poles)
        # then derive the latitude using x and y:
        normalized_z = z__ / radius
        lat_mask_cond = abs(normalized_z) < thr
        lat_z_only = 90 - np.rad2deg(np.arccos(normalized_z))
        lats = np.where(lat_mask_cond, lat_z_only, lats)

    return lons, lats


def _work_with_lonlats(klass):
    """Adapt MultipleInterpolator classes to work with geographical coordinates."""

    class GeoKlass(klass):

        def __init__(self, tie_points, *data, **interpolator_init_kwargs):
            """Set up the interpolator."""
            data = to_xyz(data)
            super().__init__(tie_points, *data, **interpolator_init_kwargs)

        def interpolate(self, fine_points, **interpolator_call_kwargs):
            """Interpolate to *fine_points*."""
            x, y, z = super().interpolate(fine_points, **interpolator_call_kwargs)
            return xyz2lonlat(x, y, z)

    return GeoKlass


def to_xyz(data):
    """Convert data to cartesian.

    Data can be a class with a `get_cartesian_coords` method, or a tuple of (lon, lat) arrays.
    """
    if len(data) == 1:
        xyz = data[0].get_cartesian_coords()
        data = [xyz[:, :, 0], xyz[:, :, 1], xyz[:, :, 2]]
    elif len(data) == 2:
        data = lonlat2xyz(*data)
    else:
        raise ValueError("Either pass lon/lats or a pyresample definition.")
    return data


GeoGridInterpolator = _work_with_lonlats(MultipleGridInterpolator)
GeoSplineInterpolator = _work_with_lonlats(MultipleSplineInterpolator)



# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False, initializedcheck=False, nonecheck=False
cimport cython

from ._modis_utils cimport floating
from ._modis_utils cimport lonlat2xyz, xyz2lonlat
from ._modis_utils import rows_per_scan_for_resolution
cimport numpy as np
import numpy as np
from scipy.ndimage import map_coordinates

np.import_array()

def interpolate_geolocation_cartesian(
        np.ndarray[floating, ndim=2] lon_array,
        np.ndarray[floating, ndim=2] lat_array,
        unsigned int coarse_resolution,
        unsigned int fine_resolution):
    lon_array = np.ascontiguousarray(lon_array)
    lat_array = np.ascontiguousarray(lat_array)
    cdef unsigned int rows_per_scan = rows_per_scan_for_resolution(coarse_resolution)
    cdef unsigned int res_factor = coarse_resolution // fine_resolution
    cdef Py_ssize_t num_rows = lon_array.shape[0]
    cdef Py_ssize_t num_cols = lon_array.shape[1]
    cdef unsigned int num_scans = num_rows // rows_per_scan

    # SciPy's map_coordinates requires the x/y dimension to be first
    cdef np.ndarray[floating, ndim=3] coordinates = np.empty(
        (2, res_factor * rows_per_scan, res_factor * num_cols), dtype=lon_array.dtype)
    cdef floating[:, :, ::1] coordinates_view = coordinates
    _compute_yx_coordinate_arrays(res_factor, coordinates_view)

    cdef np.ndarray[floating, ndim=3] xyz_result = np.empty(
        (res_factor * rows_per_scan, num_cols * res_factor, 3), dtype=lon_array.dtype)
    cdef floating[:, :, ::1] xyz_result_view = xyz_result
    cdef np.ndarray[floating, ndim=3] xyz_in = np.empty(
        (rows_per_scan, num_cols, 3), dtype=lon_array.dtype)
    cdef floating[:, :, ::1] xyz_in_view = xyz_in
    cdef floating[:, ::1] lon_in_view = lon_array
    cdef floating[:, ::1] lat_in_view = lat_array

    cdef np.ndarray[floating, ndim=2] new_lons = np.empty((res_factor * num_rows, res_factor * num_cols),
                                                          dtype=lon_array.dtype)
    cdef np.ndarray[floating, ndim=2] new_lats = np.empty((res_factor * num_rows, res_factor * num_cols),
                                                          dtype=lon_array.dtype)
    cdef floating[:, ::1] new_lons_view = new_lons
    cdef floating[:, ::1] new_lats_view = new_lats

    # Interpolate each scan, one at a time, otherwise the math doesn't work well
    cdef Py_ssize_t scan_idx, j0, j1, k0, k1, comp_index
    with nogil:
        for scan_idx in range(num_scans):
            # Calculate indexes
            j0 = rows_per_scan * scan_idx
            j1 = j0 + rows_per_scan
            k0 = rows_per_scan * res_factor * scan_idx
            k1 = k0 + rows_per_scan * res_factor
            lonlat2xyz(lon_in_view[j0:j1, :], lat_in_view[j0:j1, :], xyz_in_view)

            _compute_interpolated_xyz_scan(
                res_factor, coordinates_view, xyz_in_view,
                xyz_result_view)

            xyz2lonlat(xyz_result_view, new_lons_view[k0:k1], new_lats_view[k0:k1], low_lat_z=True)
    return new_lons, new_lats


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void _compute_yx_coordinate_arrays(
        unsigned int res_factor,
        floating[:, :, ::1] coordinates,
) noexcept nogil:
    cdef Py_ssize_t i, j
    for j in range(coordinates.shape[1]):
        for i in range(coordinates.shape[2]):
            # y coordinate - 0.375 for 250m, 0.25 for 500m
            coordinates[0, j, i] = j * (1.0 / res_factor) - (res_factor * (1.0 / 16) + (1.0 / 8))
            # x coordinate
            coordinates[1, j, i] = i * (1.0 / res_factor)


@cython.boundscheck(False)
cdef void _compute_interpolated_xyz_scan(
        unsigned int res_factor,
        floating[:, :, ::1] coordinates_view,
        floating[:, :, ::1] xyz_input_view,
        floating[:, :, ::1] xyz_result_view,
) noexcept nogil:
    cdef Py_ssize_t comp_index
    cdef floating[:, :] input_view, result_view
    with gil:
        for comp_index in range(3):
            input_view = xyz_input_view[:, :, comp_index]
            result_view = xyz_result_view[:, :, comp_index]
            _call_map_coordinates(
                input_view,
                coordinates_view,
                result_view,
            )

    if res_factor == 4:
        for comp_index in range(3):
            result_view = xyz_result_view[:, :, comp_index]
            _extrapolate_xyz_rightmost_columns(result_view, 3)
            _interpolate_xyz_250(
                result_view,
                coordinates_view,
            )
    else:
        for comp_index in range(3):
            result_view = xyz_result_view[:, :, comp_index]
            _extrapolate_xyz_rightmost_columns(result_view, 1)
            _interpolate_xyz_500(
                result_view,
                coordinates_view,
            )


cdef void _call_map_coordinates(
        floating[:, :] nav_array_view,
        floating[:, :, ::1] coordinates_view,
        floating[:, :] result_view,
):
    cdef np.ndarray[floating, ndim=2] nav_array = np.asarray(nav_array_view)
    cdef np.ndarray[floating, ndim=3] coordinates_array = np.asarray(coordinates_view)
    cdef np.ndarray[floating, ndim=2] result_array = np.asarray(result_view)
    # Use bilinear interpolation for all 250 meter pixels
    map_coordinates(nav_array, coordinates_array,
                    output=result_array,
                    order=1, mode='nearest')


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void _extrapolate_xyz_rightmost_columns(
        floating[:, :] result_view,
        int num_columns,
) noexcept nogil:
    cdef Py_ssize_t row_idx, col_offset
    cdef floating last_interp_col_diff
    for row_idx in range(result_view.shape[0]):
        last_interp_col_diff = result_view[row_idx, result_view.shape[1] - num_columns - 1] - \
                               result_view[row_idx, result_view.shape[1] - num_columns - 2]
        for col_offset in range(num_columns):
            # map_coordinates repeated the last columns value, we now add more to it as an "extrapolation"
            result_view[row_idx, result_view.shape[1] - num_columns + col_offset] += last_interp_col_diff * (col_offset + 1)


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void _interpolate_xyz_250(
        floating[:, :] result_view,
        floating[:, :, ::1] coordinates_view,
) noexcept nogil:
    cdef Py_ssize_t col_idx
    cdef floating m, b
    cdef floating[:] result_col_view
    cdef floating[:, ::1] y_coordinates = coordinates_view[0]
    for col_idx in range(result_view.shape[1]):
        result_col_view = result_view[:, col_idx]
        # Use linear extrapolation for the first two 250 meter pixels along track
        m = _calc_slope_250(result_col_view,
                            y_coordinates,
                            2)
        b = _calc_offset_250(result_col_view,
                             y_coordinates,
                             m,
                             2)
        result_view[0, col_idx] = m * y_coordinates[0, 0] + b
        result_view[1, col_idx] = m * y_coordinates[1, 0] + b

        # Use linear extrapolation for the last  two 250 meter pixels along track
        # m = (result_array[k0 + 37, :] - result_array[k0 + 34, :]) / (y[37, 0] - y[34, 0])
        # b = result_array[k0 + 37, :] - m * y[37, 0]
        m = _calc_slope_250(result_col_view,
                            y_coordinates,
                            34)
        b = _calc_offset_250(result_col_view,
                             y_coordinates,
                             m,
                             34)
        result_view[38, col_idx] = m * y_coordinates[38, 0] + b
        result_view[39, col_idx] = m * y_coordinates[39, 0] + b


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void _interpolate_xyz_500(
        floating[:, :] result_view,
        floating[:, :, ::1] coordinates_view,
) noexcept nogil:
    cdef Py_ssize_t col_idx
    cdef floating m, b
    for col_idx in range(result_view.shape[1]):
        # Use linear extrapolation for the first two 250 meter pixels along track
        m = _calc_slope_500(
            result_view[:, col_idx],
            coordinates_view[0],
            1)
        b = _calc_offset_500(
            result_view[:, col_idx],
            coordinates_view[0],
            m,
            1)
        result_view[0, col_idx] = m * coordinates_view[0, 0, 0] + b

        # Use linear extrapolation for the last two 250 meter pixels along track
        m = _calc_slope_500(
            result_view[:, col_idx],
            coordinates_view[0],
            17)
        b = _calc_offset_500(
            result_view[:, col_idx],
            coordinates_view[0],
            m,
            17)
        result_view[19, col_idx] = m * coordinates_view[0, 19, 0] + b


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline floating _calc_slope_250(
        floating[:] result_view,
        floating[:, ::1] y,
        Py_ssize_t offset,
) noexcept nogil:
    return (result_view[offset + 3] - result_view[offset]) / \
           (y[offset + 3, 0] - y[offset, 0])


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline floating _calc_offset_250(
        floating[:] result_view,
        floating[:, ::1] y,
        floating m,
        Py_ssize_t offset,
) noexcept nogil:
    return result_view[offset + 3] - m * y[offset + 3, 0]


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline floating _calc_slope_500(
        floating[:] result_view,
        floating[:, ::1] y,
        Py_ssize_t offset,
) noexcept nogil:
    return (result_view[offset + 1] - result_view[offset]) / \
           (y[offset + 1, 0] - y[offset, 0])


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline floating _calc_offset_500(
        floating[:] result_view,
        floating[:, ::1] y,
        floating m,
        Py_ssize_t offset,
) noexcept nogil:
    return result_view[offset + 1] - m * y[offset + 1, 0]



# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False, initializedcheck=False, nonecheck=False
cimport cython
import numpy as np
cimport numpy as np

np.import_array()

def test_func():
    cdef np.ndarray[float, ndim=2] arr = np.zeros((5, 5), dtype=np.float32)
    cdef float[:, ::1] arr_view = arr
    t = Test(5.0)
    t.call_me(arr_view)


cdef class Test:

    cdef float _a

    def __cinit__(self, float a):
        self._a = a

    cdef void call_me(self, float[:, ::1] my_arr) noexcept:
        with nogil:
            self._call_me(my_arr)

    cdef void _call_me(self, float[:, ::1] my_arr) noexcept nogil:
        cdef Py_ssize_t idx
        cdef float[:, :] my_arr2 = _get_upper_left_corner(my_arr)
        for idx in range(my_arr2.shape[0]):
            my_arr2[idx, 0] = self._a


cdef inline float[:, :] _get_upper_left_corner(float[:, ::1] arr) noexcept nogil:
    return arr[:3, :3]



# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False, initializedcheck=False, nonecheck=False
from functools import wraps

cimport cython
cimport numpy as np
from libc.math cimport asin, sin, cos, sqrt, acos, M_PI
import numpy as np

np.import_array()

try:
    import dask.array as da
except ImportError:
    # if dask can't be imported then we aren't going to be given dask arrays
    da = None

try:
    import xarray as xr
except ImportError:
    xr = None


DEF EARTH_RADIUS = 6370997.0


cdef void lonlat2xyz(
        floating[:, ::1] lons,
        floating[:, ::1] lats,
        floating[:, :, ::1] xyz,
) noexcept nogil:
    """Convert lons and lats to cartesian coordinates."""
    cdef Py_ssize_t i, j, k
    cdef floating lon_rad, lat_rad
    for i in range(lons.shape[0]):
        for j in range(lons.shape[1]):
            lon_rad = deg2rad(lons[i, j])
            lat_rad = deg2rad(lats[i, j])
            xyz[i, j, 0] = EARTH_RADIUS * cos(lat_rad) * cos(lon_rad)
            xyz[i, j, 1] = EARTH_RADIUS * cos(lat_rad) * sin(lon_rad)
            xyz[i, j, 2] = EARTH_RADIUS * sin(lat_rad)


cdef void xyz2lonlat(
        floating[:, :, ::1] xyz,
        floating[:, ::1] lons,
        floating[:, ::1] lats,
        bint low_lat_z=True,
        floating thr=0.8) noexcept nogil:
    """Get longitudes from cartesian coordinates."""
    cdef Py_ssize_t i, j
    cdef np.float64_t x, y, z
    for i in range(xyz.shape[0]):
        for j in range(xyz.shape[1]):
            # 64-bit precision matters apparently
            x = <np.float64_t>xyz[i, j, 0]
            y = <np.float64_t>xyz[i, j, 1]
            z = <np.float64_t>xyz[i, j, 2]
            lons[i, j] = rad2deg(acos(x / sqrt(x ** 2 + y ** 2))) * _sign(y)
            # if we are at low latitudes - small z, then get the
            # latitudes only from z. If we are at high latitudes (close to the poles)
            # then derive the latitude using x and y:
            if low_lat_z and (z < thr * EARTH_RADIUS) and (z > -1.0 * thr * EARTH_RADIUS):
                lats[i, j] = 90 - rad2deg(acos(z / EARTH_RADIUS))
            else:
                lats[i, j] = _sign(z) * (90 - rad2deg(asin(sqrt(x ** 2 + y ** 2) / EARTH_RADIUS)))


cdef inline int _sign(floating x) noexcept nogil:
    return 1 if x > 0 else (-1 if x < 0 else 0)


cdef inline floating rad2deg(floating x) noexcept nogil:
    return x * (180.0 / M_PI)


cdef inline floating deg2rad(floating x) noexcept nogil:
    return x * (M_PI / 180.0)


def rows_per_scan_for_resolution(res):
    return {
        5000: 2,
        1000: 10,
        500: 20,
        250: 40,
    }[res]


def scanline_mapblocks(func):
    """Convert dask array inputs to appropriate map_blocks calls.

    This function, applied as a decorator, will call the wrapped function
    using dask's ``map_blocks``. It will rechunk dask array inputs when
    necessary to make sure that the input chunks are entire scanlines to
    avoid incorrect interpolation.

    """
    @wraps(func)
    def _wrapper(*args, coarse_resolution=None, fine_resolution=None, **kwargs):
        if coarse_resolution is None or fine_resolution is None:
            raise ValueError("'coarse_resolution' and 'fine_resolution' are required keyword arguments.")
        first_arr = [arr for arr in args if hasattr(arr, "ndim")][0]
        if first_arr.ndim != 2 or first_arr.ndim != 2:
            raise ValueError("Expected 2D input arrays.")
        if hasattr(first_arr, "compute"):
            # assume it is dask or xarray with dask, ensure proper chunk size
            # if DataArray get just the dask array
            dask_args = _extract_dask_arrays_from_args(args)
            rows_per_scan = rows_per_scan_for_resolution(coarse_resolution)
            rechunked_args = _rechunk_dask_arrays_if_needed(dask_args, rows_per_scan)
            results = _call_map_blocks_interp(
                func,
                coarse_resolution,
                fine_resolution,
                *rechunked_args,
                **kwargs
            )
            if hasattr(first_arr, "dims"):
                # recreate DataArrays
                results = _results_to_data_arrays(first_arr.dims, *results)
            return results
        return func(
            *args,
            coarse_resolution=coarse_resolution,
            fine_resolution=fine_resolution,
            **kwargs
        )

    return _wrapper


def _extract_dask_arrays_from_args(args):
    return [arr_obj.data if hasattr(arr_obj, "dims") else arr_obj for arr_obj in args]


def _call_map_blocks_interp(func, coarse_resolution, fine_resolution, *args, **kwargs):
    first_arr = [arr for arr in args if hasattr(arr, "ndim")][0]
    res_factor = coarse_resolution // fine_resolution
    new_row_chunks = tuple(x * res_factor for x in first_arr.chunks[0])
    fine_pixels_per_1km = {250: 4, 500: 2, 1000: 1}[fine_resolution]
    fine_scan_width = 1354 * fine_pixels_per_1km
    new_col_chunks = (fine_scan_width,)
    wrapped_func = _map_blocks_handler(func)
    res = da.map_blocks(wrapped_func, *args,
                        coarse_resolution=coarse_resolution,
                        fine_resolution=fine_resolution,
                        **kwargs,
                        new_axis=[0],
                        chunks=(2, new_row_chunks, new_col_chunks),
                        dtype=first_arr.dtype,
                        meta=np.empty((2, 2, 2), dtype=first_arr.dtype))
    return tuple(res[idx] for idx in range(res.shape[0]))


def _results_to_data_arrays(dims, *results):
    new_results = []
    for result in results:
        if not isinstance(result, da.Array):
            continue
        data_arr = xr.DataArray(result, dims=dims)
        new_results.append(data_arr)
    return new_results


def _rechunk_dask_arrays_if_needed(args, rows_per_scan: int):
    # take current chunk size and get a relatively similar chunk size
    first_arr = [arr for arr in args if hasattr(arr, "ndim")][0]
    row_chunks = first_arr.chunks[0]
    col_chunks = first_arr.chunks[1]
    num_rows = first_arr.shape[0]
    num_cols = first_arr.shape[1]
    good_row_chunks = all(x % rows_per_scan == 0 for x in row_chunks)
    good_col_chunks = len(col_chunks) == 1 and col_chunks[0] != num_cols
    all_orig_chunks = [arr.chunks for arr in args if hasattr(arr, "chunks")]

    if num_rows % rows_per_scan != 0:
        raise ValueError("Input longitude/latitude data does not consist of "
                         "whole scans (10 rows per scan).")
    all_same_chunks = all(
        all_orig_chunks[0] == some_chunks
        for some_chunks in all_orig_chunks[1:]
    )
    if good_row_chunks and good_col_chunks and all_same_chunks:
        return args

    new_row_chunks = (row_chunks[0] // rows_per_scan) * rows_per_scan
    new_args = [arr.rechunk((new_row_chunks, -1)) if hasattr(arr, "chunks") else arr for arr in args]
    return new_args


def _map_blocks_handler(func):
    @wraps(func)
    def _map_blocks_wrapper(*args, **kwargs):
        results = func(*args, **kwargs)
        return np.concatenate(
            tuple(result[np.newaxis] for result in results),
            axis=0)
    return _map_blocks_wrapper





"""Test of the interpolation of geographical tiepoints for the VII products.

It follows the description provided in document "EPS-SG VII Level 1B Product Format Specification V4A".
This version is compatible for vii (METimage) test data version V2 (Jan 2022). It is not back compatible
with V1.

"""

import unittest
import numpy as np
import xarray as xr
import pytest
from geotiepoints.viiinterpolator import tie_points_interpolation, tie_points_geo_interpolation


TEST_N_SCANS = 2
TEST_TIE_POINTS_FACTOR = 2
TEST_SCAN_ALT_TIE_POINTS = 3
TEST_VALID_ALT_TIE_POINTS = TEST_SCAN_ALT_TIE_POINTS * TEST_N_SCANS
TEST_INVALID_ALT_TIE_POINTS = TEST_SCAN_ALT_TIE_POINTS * TEST_N_SCANS + 1
TEST_ACT_TIE_POINTS = 4

# Results of latitude/longitude interpolation with simple interpolation on coordinates
TEST_LON_1 = np.array(
    [[-12., -11.5, -11., -10.5, -10.0, -9.5],
     [-10., -9.5, -9., -8.5, -8.0, -7.5],
     [-8., -7.5, -7., -6.5, -6.0, -5.5],
     [-6., -5.5, -5., -4.5, -4.0, -3.5],
     [0., 0.5, 1., 1.5, 2.0, 2.5],
     [2., 2.5, 3., 3.5, 4.0, 4.5],
     [4., 4.5, 5., 5.5, 6.0, 6.5],
     [6., 6.5, 7., 7.5, 8.0, 8.5]]
)
TEST_LAT_1 = np.array(
    [[0., 0.5, 1., 1.5, 2., 2.5],
     [2., 2.5, 3., 3.5, 4., 4.5],
     [4., 4.5, 5., 5.5, 6., 6.5],
     [6., 6.5, 7., 7.5, 8., 8.5],
     [12., 12.5, 13., 13.5, 14.0, 14.5],
     [14., 14.5, 15., 15.5, 16., 16.5],
     [16., 16.5, 17., 17.5, 18., 18.5],
     [18., 18.5, 19., 19.5, 20., 20.5]]
)

# Results of latitude/longitude interpolation on cartesian coordinates (latitude above 60 degrees)
TEST_LON_2 = np.array(
    [[-12., -11.50003808, -11., -10.50011426, -10., -9.50019052],
     [-10.00243991, -9.5032411, -9.00366173, -8.50454031, -8.00488578, -7.5058423],
     [-8., -7.50034342, -7., -6.50042016, -6., -5.50049716],
     [-6.00734362, -5.50845783, -5.00857895, -4.50977302, -4.00981958, -3.51109426],
     [0., 0.49903263, 1., 1.49895241, 2., 2.49887151],
     [1.98257947, 2.48080192, 2.98127841, 3.47941324, 3.97996512, 4.47801105],
     [4., 4.49870746, 5., 5.49862418, 6., 6.49853998],
     [5.97729789, 6.47516183, 6.97594186, 7.47371256, 7.97456943, 8.47224525]]
)

TEST_LAT_2 = np.array(
    [[0., 0.49998096, 1., 1.49994287, 2., 2.49990475],
     [1.99878116, 2.49838091, 2.99817081, 3.49773189, 3.99755935, 4.49708148],
     [4., 4.4998283, 5., 5.49978993, 6., 6.49975143],
     [5.99633155, 6.4957749, 6.99571447, 7.49511791, 7.99509473, 8.49445789],
     [12., 12.49951634, 13., 13.49947623, 14., 14.499435779],
     [13.99129786, 14.4904098, 14.99064796, 15.48971613, 15.98999196, 16.48901572],
     [16., 16.49935377, 17., 17.49931213, 18., 18.49927003],
     [17.98865968, 18.48759253, 18.98798235, 19.48686863, 19.98729684, 20.48613573]]
)

# Results of latitude/longitude interpolation on cartesian coordinates (longitude with a 360 degrees step)
TEST_LON_3 = np.array(
    [[-12., -11.50444038, -11., -10.50459822, -10., -9.50476197],
     [-10.07492627, -9.58101155, -9.07759836, -8.5839056, - 8.0803761, -7.58691614],
     [-8., -7.50510905, -7., -6.5052934, -6., -5.50548573],
     [-6.0862821, -5.59332416, -5.08942935, -4.59674283, -4.09272043, -3.60032066],
     [0., 0.49315061, 1., 1.49287934, 2., 2.49259217],
     [1.88371709, 2.3739768, 2.87898193, 3.36879304, 3.87395153, 4.36327924],
     [4., 4.49196335, 5., 5.49161771, 6., 6.49124808],
     [5.86287282, 6.35111105, 6.85674573, 7.3443667, 7.85016382, 8.33710998]]
)

TEST_LAT_3 = np.array(
    [[45., 45.49777998, 46., 46.49770107, 47., 47.4976192],
     [46.96258417, 47.4595462, 47.9612508, 48.4581022, 48.95986481, 49.45660021],
     [49., 49.49744569, 50., 50.49735352, 51., 51.49725738],
     [50.95691833, 51.45340364, 51.95534841, 52.45169855, 52.95370691, 53.55477452],
     [57., 57.50277937, 58., 58.50267347, 59., 59.50256981],
     [59.04185095, 59.54357898, 60.04020844, 60.54185139, 61.03859839, 61.54015723],
     [61., 61.5023687, 62., 62.502271, 63., 63.50217506],
     [63.03546804, 63.53686131, 64.03394419, 64.53525587, 65.03244567, 65.53367647]]
)


class TestViiInterpolator(unittest.TestCase):
    """Test the vii_utils module."""

    def setUp(self):
        """Set up the test."""
        # Create the arrays for the interpolation test
        # The first has a valid number of n_tie_alt points (multiple of SCAN_ALT_TIE_POINTS)
        self.valid_data_for_interpolation = xr.DataArray(
            np.arange(
                TEST_VALID_ALT_TIE_POINTS * TEST_ACT_TIE_POINTS,
                dtype=np.float64,
            ).reshape(TEST_VALID_ALT_TIE_POINTS, TEST_ACT_TIE_POINTS),
            dims=('num_tie_points_alt', 'num_tie_points_act'),
        )
        # The second has an invalid number of n_tie_alt points (not multiple of SCAN_ALT_TIE_POINTS)
        self.invalid_data_for_interpolation = xr.DataArray(
            np.arange(
                TEST_INVALID_ALT_TIE_POINTS * TEST_ACT_TIE_POINTS,
                dtype=np.float64,
            ).reshape(TEST_INVALID_ALT_TIE_POINTS, TEST_ACT_TIE_POINTS),
            dims=('num_tie_points_alt', 'num_tie_points_act'),
        )
        # Then two arrays containing valid longitude and latitude data
        self.longitude = xr.DataArray(
            np.linspace(
                -12,
                11,
                num=TEST_VALID_ALT_TIE_POINTS * TEST_ACT_TIE_POINTS,
                dtype=np.float64,
            ).reshape(TEST_VALID_ALT_TIE_POINTS, TEST_ACT_TIE_POINTS),
            dims=('num_tie_points_alt', 'num_tie_points_act'),
        )
        self.latitude = xr.DataArray(
            np.linspace(
                0,
                23,
                num=TEST_VALID_ALT_TIE_POINTS * TEST_ACT_TIE_POINTS,
                dtype=np.float64,
            ).reshape(TEST_VALID_ALT_TIE_POINTS, TEST_ACT_TIE_POINTS),
            dims=('num_tie_points_alt', 'num_tie_points_act'),
        )
        # Then one containing latitude data above 60 degrees
        self.latitude_over60 = xr.DataArray(
            np.linspace(
                45,
                68,
                num=TEST_VALID_ALT_TIE_POINTS * TEST_ACT_TIE_POINTS,
                dtype=np.float64,
            ).reshape(TEST_VALID_ALT_TIE_POINTS, TEST_ACT_TIE_POINTS),
            dims=('num_tie_points_alt', 'num_tie_points_act'),
        )
        # Then one containing longitude data with a 360 degrees step
        self.longitude_over360 = xr.DataArray(
            np.linspace(
                -12,
                11,
                num=TEST_VALID_ALT_TIE_POINTS * TEST_ACT_TIE_POINTS,
                dtype=np.float64,
            ).reshape(TEST_VALID_ALT_TIE_POINTS, TEST_ACT_TIE_POINTS) % 360.,
            dims=('num_tie_points_alt', 'num_tie_points_act'),
        )

    def tearDown(self):
        """Tear down the test."""
        # Nothing to do
        pass

    def test_tie_points_interpolation(self):
        """# Test the interpolation routine with valid and invalid input."""
        # Test the interpolation routine with valid input
        result_valid = tie_points_interpolation(
            [self.valid_data_for_interpolation],
            TEST_SCAN_ALT_TIE_POINTS,
            TEST_TIE_POINTS_FACTOR
        )[0]

        act_points_interp = (TEST_ACT_TIE_POINTS - 1) * TEST_TIE_POINTS_FACTOR
        num_scans = TEST_VALID_ALT_TIE_POINTS // TEST_SCAN_ALT_TIE_POINTS
        scan_alt_points_interp = (TEST_SCAN_ALT_TIE_POINTS - 1) * TEST_TIE_POINTS_FACTOR

        # Across the track
        delta_axis_0 = [0., 0.5, 1., 1.5, 2., 2.5]
        self.assertTrue(np.allclose(result_valid[0, :], delta_axis_0))
        # Along track
        delta_axis_1 = [0., 2., 4., 6., 12., 14., 16., 18]
        self.assertTrue(np.allclose(result_valid[:, 0], delta_axis_1))

        # Test the interpolation routine with invalid input
        pytest.raises(ValueError, tie_points_interpolation,
                      [self.invalid_data_for_interpolation],
                      TEST_SCAN_ALT_TIE_POINTS,
                      TEST_TIE_POINTS_FACTOR)

    def test_tie_points_geo_interpolation(self):
        """# Test the coordinates interpolation routine with valid and invalid input."""
        # Test the interpolation routine with valid input
        lon, lat = tie_points_geo_interpolation(
            self.longitude,
            self.latitude,
            TEST_SCAN_ALT_TIE_POINTS,
            TEST_TIE_POINTS_FACTOR
        )
        self.assertTrue(np.allclose(lon, TEST_LON_1))
        self.assertTrue(np.allclose(lat, TEST_LAT_1))

        lon, lat = tie_points_geo_interpolation(
            self.longitude_over360,
            self.latitude,
            TEST_SCAN_ALT_TIE_POINTS,
            TEST_TIE_POINTS_FACTOR
        )
        self.assertTrue(np.allclose(lon, TEST_LON_2))
        self.assertTrue(np.allclose(lat, TEST_LAT_2))

        lon, lat = tie_points_geo_interpolation(
            self.longitude,
            self.latitude_over60,
            TEST_SCAN_ALT_TIE_POINTS,
            TEST_TIE_POINTS_FACTOR
        )
        self.assertTrue(np.allclose(lon, TEST_LON_3))
        self.assertTrue(np.allclose(lat, TEST_LAT_3))

        # Test the interpolation routine with invalid input (different dimensions of the two arrays)
        with self.assertRaises(ValueError):
            tie_points_geo_interpolation(
                self.longitude,
                self.invalid_data_for_interpolation,
                TEST_SCAN_ALT_TIE_POINTS,
                TEST_TIE_POINTS_FACTOR
            )



"""Interpolation of geographical tiepoints."""

from multiprocessing import Pool

import numpy as np
from numpy import arccos, arcsin, rad2deg, sign, sqrt
from scipy.interpolate import RectBivariateSpline, splev, splrep

from geotiepoints.geointerpolator import \
    GeoInterpolator as SatelliteInterpolator

EARTH_RADIUS = 6370997.0


def get_scene_splits(nlines_swath, nlines_scan, n_cpus):
    """Calculate the line numbers where the swath will be split in smaller
    granules for parallel processing"""

    nscans = nlines_swath // nlines_scan
    if nscans < n_cpus:
        nscans_subscene = 1
    else:
        nscans_subscene = nscans // n_cpus
    nlines_subscene = nscans_subscene * nlines_scan

    return range(nlines_subscene, nlines_swath, nlines_subscene)


def metop20kmto1km(lons20km, lats20km):
    """Getting 1km geolocation for metop avhrr from 20km tiepoints.
    """
    cols20km = np.array([0] + list(range(4, 2048, 20)) + [2047])
    cols1km = np.arange(2048)
    lines = lons20km.shape[0]
    rows20km = np.arange(lines)
    rows1km = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3

    satint = SatelliteInterpolator((lons20km, lats20km),
                                   (rows20km, cols20km),
                                   (rows1km, cols1km),
                                   along_track_order,
                                   cross_track_order)
    return satint.interpolate()


def modis5kmto1km(lons5km, lats5km):
    """Getting 1km geolocation for modis from 5km tiepoints.

    http://www.icare.univ-lille1.fr/tutorials/MODIS_geolocation
    """
    cols5km = np.arange(2, 1354, 5) / 5.0
    cols1km = np.arange(1354) / 5.0
    lines = lons5km.shape[0] * 5
    rows5km = np.arange(2, lines, 5) / 5.0
    rows1km = np.arange(lines) / 5.0

    along_track_order = 1
    cross_track_order = 3

    satint = SatelliteInterpolator((lons5km, lats5km),
                                   (rows5km, cols5km),
                                   (rows1km, cols1km),
                                   along_track_order,
                                   cross_track_order,
                                   chunk_size=10)
    satint.fill_borders("y", "x")
    lons1km, lats1km = satint.interpolate()
    return lons1km, lats1km


def _multi(fun, lons, lats, chunk_size, cores=1):
    """Work on multiple cores.
    """
    pool = Pool(processes=cores)

    splits = get_scene_splits(lons.shape[0], chunk_size, cores)

    lons_parts = np.vsplit(lons, splits)
    lats_parts = np.vsplit(lats, splits)

    results = [pool.apply_async(fun,
                                (lons_parts[i],
                                 lats_parts[i]))
               for i in range(len(lons_parts))]

    pool.close()
    pool.join()

    lons, lats = zip(*(res.get() for res in results))

    return np.vstack(lons), np.vstack(lats)


def modis1kmto500m(lons1km, lats1km, cores=1):
    """Getting 500m geolocation for modis from 1km tiepoints.

    http://www.icare.univ-lille1.fr/tutorials/MODIS_geolocation
    """
    if cores > 1:
        return _multi(modis1kmto500m, lons1km, lats1km, 10, cores)

    cols1km = np.arange(1354)
    cols500m = np.arange(1354 * 2) / 2.0
    lines = lons1km.shape[0]
    rows1km = np.arange(lines)
    rows500m = (np.arange(lines * 2) - 0.5) / 2.

    along_track_order = 1
    cross_track_order = 3

    satint = SatelliteInterpolator((lons1km, lats1km),
                                   (rows1km, cols1km),
                                   (rows500m, cols500m),
                                   along_track_order,
                                   cross_track_order,
                                   chunk_size=20)
    satint.fill_borders("y", "x")
    lons500m, lats500m = satint.interpolate()
    return lons500m, lats500m


def modis1kmto250m(lons1km, lats1km, cores=1):
    """Getting 250m geolocation for modis from 1km tiepoints.

    http://www.icare.univ-lille1.fr/tutorials/MODIS_geolocation
    """
    if cores > 1:
        return _multi(modis1kmto250m, lons1km, lats1km, 10, cores)

    cols1km = np.arange(1354)
    cols250m = np.arange(1354 * 4) / 4.0

    along_track_order = 1
    cross_track_order = 3

    lines = lons1km.shape[0]
    rows1km = np.arange(lines)
    rows250m = (np.arange(lines * 4) - 1.5) / 4.0

    satint = SatelliteInterpolator((lons1km, lats1km),
                                   (rows1km, cols1km),
                                   (rows250m, cols250m),
                                   along_track_order,
                                   cross_track_order,
                                   chunk_size=40)
    satint.fill_borders("y", "x")
    lons250m, lats250m = satint.interpolate()

    return lons250m, lats250m


from . import version
__version__ = version.get_versions()['version']




# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False, initializedcheck=False, nonecheck=False
cimport cython
from ._modis_utils cimport lonlat2xyz, xyz2lonlat, floating, deg2rad
from .simple_modis_interpolator import scanline_mapblocks

from libc.math cimport asin, sin, cos, sqrt
cimport numpy as np
import numpy as np

DEF R = 6370.997
# Aqua altitude in km
DEF H = 709.0

np.import_array()


@scanline_mapblocks
def interpolate(
        np.ndarray[floating, ndim=2] lon1,
        np.ndarray[floating, ndim=2] lat1,
        np.ndarray[floating, ndim=2] satz1,
        unsigned int coarse_resolution=0,
        unsigned int fine_resolution=0,
        unsigned int coarse_scan_width=0,
):
    """Helper function to interpolate scan-aligned arrays.

    This function's decorator runs this function for each dask block/chunk of
    scans. The arrays are scan-aligned meaning they are an even number of scans
    (N rows per scan) and contain the entire scan width.

    """
    if coarse_resolution == 5000 and coarse_scan_width not in (0, 270, 271):
        raise NotImplementedError(
            "Can't interpolate if 5km tiepoints have less than 270 columns."
        )
    interp = MODISInterpolator(coarse_resolution, fine_resolution, coarse_scan_width=coarse_scan_width or 0)
    return interp.interpolate(lon1, lat1, satz1)


cdef void _compute_expansion_alignment(floating[:, :] satz_a, floating [:, :] satz_b, int scan_width,
                                       floating[:, ::1] c_expansion, floating[:, ::1] c_alignment) noexcept nogil:
    """Fill in expansion and alignment.
    
    Input angles should be in degrees and will be converted to radians.
    
    """
    cdef Py_ssize_t i, j
    cdef floating satz_a_rad, satz_b_rad, phi_a, phi_b, theta_a, theta_b, phi, zeta, theta, denominator, sin_beta_2, d, e
    for i in range(satz_a.shape[0]):
        for j in range(satz_a.shape[1]):
            satz_a_rad = deg2rad(satz_a[i, j])
            satz_b_rad = deg2rad(satz_b[i, j])
            phi_a = _compute_phi(satz_a_rad)
            phi_b = _compute_phi(satz_b_rad)
            theta_a = _compute_theta(satz_a_rad, phi_a)
            theta_b = _compute_theta(satz_b_rad, phi_b)
            phi = (phi_a + phi_b) / 2
            zeta = _compute_zeta(phi)
            theta = _compute_theta(zeta, phi)
            # Workaround for tiepoints symmetrical about the subsatellite-track
            denominator = theta_a * 2 if theta_a == theta_b else theta_a - theta_b

            c_expansion[i, j] = 4 * (((theta_a + theta_b) / 2 - theta) / denominator)

            sin_beta_2 = scan_width / (2 * H)
            d = ((R + H) / R * cos(phi) - cos(zeta)) * sin_beta_2
            e = cos(zeta) - sqrt(cos(zeta) ** 2 - d ** 2)

            c_alignment[i, j] = 4 * e * sin(zeta) / denominator


cdef inline floating _compute_phi(floating zeta) noexcept nogil:
    return asin(R * sin(zeta) / (R + H))


cdef inline floating _compute_theta(floating zeta, floating phi) noexcept nogil:
    return zeta - phi


cdef inline floating _compute_zeta(floating phi) noexcept nogil:
    return asin((R + H) * sin(phi) / R)


cdef inline floating[:, :] _get_upper_left_corner(floating[:, ::1] arr) noexcept nogil:
    return arr[:arr.shape[0] - 1, :arr.shape[1] - 1]


cdef inline floating[:, :] _get_upper_right_corner(floating[:, ::1] arr) noexcept nogil:
    return arr[:arr.shape[0] - 1, 1:]


cdef class MODISInterpolator:
    """Helper class for MODIS interpolation.

    Not intended for public use. Use ``modis_X_to_Y`` functions instead.

    """

    cdef int _coarse_scan_length
    cdef int _coarse_scan_width
    cdef int _coarse_pixels_per_1km
    cdef int _fine_pixels_per_coarse_pixel
    cdef int _fine_scan_width
    cdef int _fine_scan_length
    cdef int _coarse_resolution
    cdef int _fine_resolution
    cdef Py_ssize_t _factor_5km

    def __cinit__(self, unsigned int coarse_resolution, unsigned int fine_resolution, unsigned int coarse_scan_width=0):
        if coarse_resolution == 1000:
            self._coarse_scan_length = 10
            self._coarse_scan_width = 1354
        elif coarse_resolution == 5000:
            self._coarse_scan_length = 2
            if coarse_scan_width == 0:
                self._coarse_scan_width = 271
            else:
                self._coarse_scan_width = coarse_scan_width
        self._coarse_pixels_per_1km = coarse_resolution // 1000

        cdef int fine_pixels_per_1km = 1000 // fine_resolution
        self._fine_pixels_per_coarse_pixel = fine_pixels_per_1km * self._coarse_pixels_per_1km
        self._fine_scan_width = 1354 * fine_pixels_per_1km
        self._coarse_resolution = coarse_resolution
        self._fine_resolution = fine_resolution
        # partial rows/columns to repeat: 5km->1km => 2, 5km->500m => 4, 5km->250m => 8
        self._factor_5km = self._fine_pixels_per_coarse_pixel // self._coarse_pixels_per_1km * 2

    cdef tuple interpolate(
            self,
            np.ndarray[floating, ndim=2] lon1,
            np.ndarray[floating, ndim=2] lat1,
            np.ndarray[floating, ndim=2] satz1):
        """Interpolate MODIS geolocation from 'coarse_resolution' to 'fine_resolution'."""
        lon1 = np.ascontiguousarray(lon1)
        lat1 = np.ascontiguousarray(lat1)
        satz1 = np.ascontiguousarray(satz1)

        cdef int num_fine_scan_rows = self._coarse_scan_length * self._fine_pixels_per_coarse_pixel
        cdef int num_fine_scan_cols = self._fine_scan_width
        cdef np.ndarray[floating, ndim=2] a_track = np.empty((num_fine_scan_rows, num_fine_scan_cols), dtype=lon1.dtype)
        cdef np.ndarray[floating, ndim=2] a_scan = np.empty((num_fine_scan_rows, num_fine_scan_cols), dtype=lon1.dtype)
        cdef floating[:, ::1] a_track_view = a_track
        cdef floating[:, ::1] a_scan_view = a_scan

        cdef tuple coords_xy = self._get_coords()
        cdef np.ndarray[floating, ndim=1] x = coords_xy[0]
        cdef np.ndarray[floating, ndim=1] y = coords_xy[1]
        cdef floating[::1] x_view = x
        cdef floating[::1] y_view = y

        cdef np.ndarray[floating, ndim=2] tmp_tiepoint_a, tmp_tiepoint_b, tmp_tiepoint_c, tmp_tiepoint_d
        cdef floating[:, ::1] tmp_tiepoint_a_view, tmp_tiepoint_b_view, tmp_tiepoint_c_view, tmp_tiepoint_d_view
        tmp_tiepoint_a = np.empty((num_fine_scan_rows, num_fine_scan_cols), dtype=lon1.dtype)
        tmp_tiepoint_b = np.empty((num_fine_scan_rows, num_fine_scan_cols), dtype=lon1.dtype)
        tmp_tiepoint_c = np.empty((num_fine_scan_rows, num_fine_scan_cols), dtype=lon1.dtype)
        tmp_tiepoint_d = np.empty((num_fine_scan_rows, num_fine_scan_cols), dtype=lon1.dtype)
        tmp_tiepoint_a_view = tmp_tiepoint_a
        tmp_tiepoint_b_view = tmp_tiepoint_b
        tmp_tiepoint_c_view = tmp_tiepoint_c
        tmp_tiepoint_d_view = tmp_tiepoint_d

        cdef floating[:, :, ::1] xyz_coarse_view = np.empty((self._coarse_scan_length, self._coarse_scan_width, 3), dtype=lon1.dtype)
        cdef floating[:, :, ::1] xyz_fine_view = np.empty((num_fine_scan_rows, num_fine_scan_cols, 3), dtype=lon1.dtype)

        cdef unsigned int scans = satz1.shape[0] // self._coarse_scan_length
        cdef np.ndarray[floating, ndim=2] new_lons = np.empty((satz1.shape[0] * self._fine_pixels_per_coarse_pixel, self._fine_scan_width), dtype=lon1.dtype)
        cdef np.ndarray[floating, ndim=2] new_lats = np.empty((satz1.shape[0] * self._fine_pixels_per_coarse_pixel, self._fine_scan_width), dtype=lon1.dtype)
        cdef floating[:, ::1] coarse_lons = lon1
        cdef floating[:, ::1] coarse_lats = lat1
        cdef floating[:, ::1] coarse_satz = satz1
        cdef floating[:, ::1] fine_lons = new_lons
        cdef floating[:, ::1] fine_lats = new_lats
        with nogil:
            self._interpolate_lons_lats(
                scans,
                coarse_lons,
                coarse_lats,
                coarse_satz,
                x_view,
                y_view,
                a_track_view,
                a_scan_view,
                tmp_tiepoint_a_view,
                tmp_tiepoint_b_view,
                tmp_tiepoint_c_view,
                tmp_tiepoint_d_view,
                xyz_coarse_view,
                xyz_fine_view,
                fine_lons,
                fine_lats,
            )
        return new_lons, new_lats

    cdef tuple _get_coords(self):
        cdef np.ndarray[np.float32_t, ndim=1] x, y
        cdef np.float32_t[::1] x_view, y_view
        if self._coarse_scan_length == 10:
            x = np.arange(self._fine_scan_width, dtype=np.float32) % self._fine_pixels_per_coarse_pixel
            y = np.empty((self._coarse_scan_length * self._fine_pixels_per_coarse_pixel,), dtype=np.float32)
            x_view = x
            y_view = y
            with nogil:
                self._get_coords_1km(x_view, y_view)
        else:
            x, y = self._get_coords_5km()
        return x, y

    cdef void _get_coords_1km(
            self,
            floating[::1] x_view,
            floating[::1] y_view,
    ) noexcept nogil:
        cdef unsigned int scan_idx
        cdef int i
        cdef unsigned int fine_idx
        cdef unsigned int half_scan_length = self._fine_pixels_per_coarse_pixel // 2
        cdef unsigned int fine_pixels_per_scan = self._coarse_scan_length * self._fine_pixels_per_coarse_pixel
        for fine_idx in range(fine_pixels_per_scan):
            if fine_idx < half_scan_length:
                y_view[fine_idx] = 0.5 + fine_idx - half_scan_length
            elif fine_idx >= fine_pixels_per_scan - half_scan_length:
                y_view[fine_idx] = (self._fine_pixels_per_coarse_pixel + 0.5) + (half_scan_length - (fine_pixels_per_scan - fine_idx))
            else:
                y_view[fine_idx] = ((fine_idx + half_scan_length) % self._fine_pixels_per_coarse_pixel) + 0.5

        for i in range(self._fine_pixels_per_coarse_pixel):
            x_view[(self._fine_scan_width - self._fine_pixels_per_coarse_pixel) + i] = self._fine_pixels_per_coarse_pixel + i

    @cython.wraparound(True)
    cdef tuple _get_coords_5km(self):
        cdef np.ndarray[np.float32_t, ndim=1] y = np.arange(self._fine_pixels_per_coarse_pixel * self._coarse_scan_length, dtype=np.float32) - 2
        cdef np.ndarray[np.float32_t, ndim=1] x = (np.arange(self._fine_scan_width, dtype=np.float32) - 2) % self._fine_pixels_per_coarse_pixel
        x[0] = -2
        x[1] = -1
        if self._coarse_scan_width == 271:
            x[-2] = 5
            x[-1] = 6
        else:
            # self._coarse_scan_width == 270
            x[-7] = 5
            x[-6] = 6
            x[-5] = 7
            x[-4] = 8
            x[-3] = 9
            x[-2] = 10
            x[-1] = 11
        return x, y

    cdef void _interpolate_lons_lats(self,
                                     unsigned int scans,
                                     floating[:, ::1] coarse_lons,
                                     floating[:, ::1] coarse_lats,
                                     floating[:, ::1] coarse_satz,
                                     floating[::1] x,
                                     floating[::1] y,
                                     floating[:, ::1] a_track,
                                     floating[:, ::1] a_scan,
                                     floating[:, ::1] tmp_fine_a,
                                     floating[:, ::1] tmp_fine_b,
                                     floating[:, ::1] tmp_fine_c,
                                     floating[:, ::1] tmp_fine_d,
                                     floating[:, :, ::1] coarse_xyz,
                                     floating[:, :, ::1] fine_xyz,
                                     floating[:, ::1] fine_lons,
                                     floating[:, ::1] fine_lats,
                                     ) noexcept nogil:
        cdef floating[:, ::1] lons_scan, lats_scan, satz_scan
        cdef floating[:, ::1] new_lons_scan, new_lats_scan
        cdef Py_ssize_t scan_idx
        for scan_idx in range(0, scans):
            lons_scan = coarse_lons[scan_idx * self._coarse_scan_length:(scan_idx + 1) * self._coarse_scan_length]
            lats_scan = coarse_lats[scan_idx * self._coarse_scan_length:(scan_idx + 1) * self._coarse_scan_length]
            satz_scan = coarse_satz[scan_idx * self._coarse_scan_length:(scan_idx + 1) * self._coarse_scan_length]
            new_lons_scan = fine_lons[scan_idx * self._coarse_scan_length * self._fine_pixels_per_coarse_pixel:(scan_idx + 1) * self._coarse_scan_length * self._fine_pixels_per_coarse_pixel]
            new_lats_scan = fine_lats[scan_idx * self._coarse_scan_length * self._fine_pixels_per_coarse_pixel:(scan_idx + 1) * self._coarse_scan_length * self._fine_pixels_per_coarse_pixel]
            self._get_atrack_ascan(
                satz_scan,
                x,
                y,
                tmp_fine_a[:self._coarse_scan_length - 1, :self._coarse_scan_width - 1],
                tmp_fine_b[:self._coarse_scan_length - 1, :self._coarse_scan_width - 1],
                tmp_fine_c,
                tmp_fine_d,
                a_track,
                a_scan)
            lonlat2xyz(lons_scan, lats_scan, coarse_xyz)
            self._compute_fine_xyz(a_track, a_scan, coarse_xyz,
                                   tmp_fine_a, tmp_fine_b, tmp_fine_c, tmp_fine_d,
                                   fine_xyz)

            xyz2lonlat(fine_xyz, new_lons_scan, new_lats_scan)

    cdef void _get_atrack_ascan(
            self,
            floating[:, ::1] satz,
            floating[::1] x,
            floating[::1] y,
            floating[:, ::1] c_exp_coarse,
            floating[:, ::1] c_ali_coarse,
            floating[:, ::1] c_exp_fine,
            floating[:, ::1] c_ali_fine,
            floating[:, ::1] a_track,
            floating[:, ::1] a_scan
    ) noexcept nogil:
        cdef floating[:, :] satz_a_view = _get_upper_left_corner(satz)
        cdef floating[:, :] satz_b_view = _get_upper_right_corner(satz)
        cdef Py_ssize_t scan_idx
        _compute_expansion_alignment(
            satz_a_view,
            satz_b_view,
            self._coarse_pixels_per_1km,
            c_exp_coarse,
            c_ali_coarse)

        cdef floating[:, :] c_exp_view2 = c_exp_coarse
        self._expand_tiepoint_array(c_exp_view2, c_exp_fine)

        cdef floating[:, :] c_ali_view2 = c_ali_coarse
        self._expand_tiepoint_array(c_ali_view2, c_ali_fine)

        self._calculate_atrack_ascan(
            x, y,
            c_exp_fine, c_ali_fine,
            a_track, a_scan)

    cdef void _calculate_atrack_ascan(
            self,
            floating[::1] coords_x,
            floating[::1] coords_y,
            floating[:, ::1] c_exp_full,
            floating[:, ::1] c_ali_full,
            floating[:, ::1] a_track,
            floating[:, ::1] a_scan,
    ) noexcept nogil:
        cdef Py_ssize_t i, j
        cdef floating s_s, s_t
        for j in range(coords_y.shape[0]):
            for i in range(coords_x.shape[0]):
                s_s = coords_x[i] / self._fine_pixels_per_coarse_pixel
                s_t = coords_y[j] / self._fine_pixels_per_coarse_pixel
                a_track[j, i] = s_t
                a_scan[j, i] = s_s + s_s * (1 - s_s) * c_exp_full[j, i] + s_t * (1 - s_t) * c_ali_full[j, i]

    cdef void _compute_fine_xyz(
            self,
            floating[:, ::1] a_track_view,
            floating[:, ::1] a_scan_view,
            floating[:, :, ::1] xyz_view,
            floating[:, ::1] data_tiepoint_a_view,
            floating[:, ::1] data_tiepoint_b_view,
            floating[:, ::1] data_tiepoint_c_view,
            floating[:, ::1] data_tiepoint_d_view,
            floating[:, :, ::1] xyz_comp_view,
    ) noexcept nogil:
        cdef Py_ssize_t k
        cdef floating[:, :] comp_a_view, comp_b_view, comp_c_view, comp_d_view
        for k in range(3):  # xyz
            comp_a_view = xyz_view[:xyz_view.shape[0] - 1, :xyz_view.shape[1] - 1, k]  # upper left
            comp_b_view = xyz_view[:xyz_view.shape[0] - 1, 1:, k]  # upper right
            comp_c_view = xyz_view[1:, 1:, k]  # lower right
            comp_d_view = xyz_view[1:, :xyz_view.shape[1] - 1, k]  # lower left
            self._expand_tiepoint_array(comp_a_view, data_tiepoint_a_view)
            self._expand_tiepoint_array(comp_b_view, data_tiepoint_b_view)
            self._expand_tiepoint_array(comp_c_view, data_tiepoint_c_view)
            self._expand_tiepoint_array(comp_d_view, data_tiepoint_d_view)
            self._compute_fine_xyz_component(
                a_track_view,
                a_scan_view,
                data_tiepoint_a_view,
                data_tiepoint_b_view,
                data_tiepoint_c_view,
                data_tiepoint_d_view,
                xyz_comp_view,
                k,
            )

    cdef void _compute_fine_xyz_component(
            self,
            floating[:, ::1] a_track_view,
            floating[:, ::1] a_scan_view,
            floating[:, ::1] data_a_2d_view,
            floating[:, ::1] data_b_2d_view,
            floating[:, ::1] data_c_2d_view,
            floating[:, ::1] data_d_2d_view,
            floating[:, :, ::1] xyz_comp_view,
            Py_ssize_t k,
    ) noexcept nogil:
        cdef Py_ssize_t i, j
        cdef floating scan1_tmp, scan2_tmp, atrack1, ascan1
        for i in range(a_scan_view.shape[0]):
            for j in range(a_scan_view.shape[1]):
                atrack1 = a_track_view[i, j]
                ascan1 = a_scan_view[i, j]
                scan1_tmp = (1 - ascan1) * data_a_2d_view[i, j] + ascan1 * data_b_2d_view[i, j]
                scan2_tmp = (1 - ascan1) * data_d_2d_view[i, j] + ascan1 * data_c_2d_view[i, j]
                xyz_comp_view[i, j, k] = (1 - atrack1) * scan1_tmp + atrack1 * scan2_tmp

    cdef void _expand_tiepoint_array(
            self,
            floating[:, :] input_arr,
            floating[:, ::1] output_arr,
    ) noexcept nogil:
        if self._coarse_scan_length == 10:
            self._expand_tiepoint_array_1km(input_arr, output_arr)
        else:
            self._expand_tiepoint_array_5km(input_arr, output_arr)

    cdef void _expand_tiepoint_array_1km(self, floating[:, :] input_arr, floating[:, ::1] expanded_arr) noexcept nogil:
        self._expand_tiepoint_array_1km_main(input_arr, expanded_arr)
        self._expand_tiepoint_array_1km_right_column(input_arr, expanded_arr)

    cdef void _expand_tiepoint_array_1km_main(self, floating[:, :] input_arr, floating[:, ::1] expanded_arr) noexcept nogil:
        cdef floating tiepoint_value
        cdef Py_ssize_t row_idx, col_idx, length_repeat_cycle, width_repeat_cycle, half_coarse_pixel_fine_offset, row_offset, col_offset
        cdef Py_ssize_t row_repeat_offset, col_repeat_offset
        half_coarse_pixel_fine_offset = self._fine_pixels_per_coarse_pixel // 2
        for row_idx in range(input_arr.shape[0]):
            row_offset = row_idx * self._fine_pixels_per_coarse_pixel + half_coarse_pixel_fine_offset
            for col_idx in range(input_arr.shape[1]):
                col_offset = col_idx * self._fine_pixels_per_coarse_pixel
                tiepoint_value = input_arr[row_idx, col_idx]
                self._expand_tiepoint_array_1km_with_repeat(tiepoint_value, expanded_arr, row_offset, col_offset)

        row_idx = 0
        row_offset = row_idx * self._fine_pixels_per_coarse_pixel
        for col_idx in range(input_arr.shape[1]):
            col_offset = col_idx * self._fine_pixels_per_coarse_pixel
            tiepoint_value = input_arr[row_idx, col_idx]
            self._expand_tiepoint_array_1km_with_repeat(tiepoint_value, expanded_arr, row_offset, col_offset)

        row_idx = input_arr.shape[0] - 1
        row_offset = row_idx * self._fine_pixels_per_coarse_pixel + self._fine_pixels_per_coarse_pixel
        for col_idx in range(input_arr.shape[1]):
            col_offset = col_idx * self._fine_pixels_per_coarse_pixel
            tiepoint_value = input_arr[row_idx, col_idx]
            self._expand_tiepoint_array_1km_with_repeat(tiepoint_value, expanded_arr, row_offset, col_offset)

    cdef void _expand_tiepoint_array_1km_right_column(
            self,
            floating[:, :] input_arr,
            floating[:, ::1] expanded_arr
    ) noexcept nogil:
        cdef floating tiepoint_value
        cdef Py_ssize_t row_idx, col_idx, length_repeat_cycle, width_repeat_cycle, half_coarse_pixel_fine_offset, row_offset, col_offset
        cdef Py_ssize_t row_repeat_offset, col_repeat_offset
        half_coarse_pixel_fine_offset = self._fine_pixels_per_coarse_pixel // 2
        col_idx = input_arr.shape[1] - 1
        col_offset = col_idx * self._fine_pixels_per_coarse_pixel + self._fine_pixels_per_coarse_pixel
        for row_idx in range(input_arr.shape[0]):
            row_offset = row_idx * self._fine_pixels_per_coarse_pixel + half_coarse_pixel_fine_offset
            tiepoint_value = input_arr[row_idx, col_idx]
            self._expand_tiepoint_array_1km_with_repeat(tiepoint_value, expanded_arr, row_offset, col_offset)

        self._expand_tiepoint_array_1km_right_column_top_row(input_arr, expanded_arr)
        self._expand_tiepoint_array_1km_right_column_bottom_row(input_arr, expanded_arr)

    cdef void _expand_tiepoint_array_1km_right_column_top_row(
            self,
            floating[:, :] input_arr,
            floating[:, ::1] expanded_arr
    ) noexcept nogil:
        cdef floating tiepoint_value
        cdef Py_ssize_t row_idx, col_idx, row_offset, col_offset
        row_idx = 0
        col_idx = input_arr.shape[1] - 1
        tiepoint_value = input_arr[row_idx, col_idx]
        row_offset = row_idx * self._fine_pixels_per_coarse_pixel
        col_offset = col_idx * self._fine_pixels_per_coarse_pixel + self._fine_pixels_per_coarse_pixel
        self._expand_tiepoint_array_1km_with_repeat(tiepoint_value, expanded_arr, row_offset, col_offset)

    cdef void _expand_tiepoint_array_1km_right_column_bottom_row(
            self,
            floating[:, :] input_arr,
            floating[:, ::1] expanded_arr
    ) noexcept nogil:
        cdef floating tiepoint_value
        cdef Py_ssize_t row_idx, col_idx, row_offset, col_offset
        row_idx = input_arr.shape[0] - 1
        col_idx = input_arr.shape[1] - 1
        tiepoint_value = input_arr[row_idx, col_idx]
        row_offset = row_idx * self._fine_pixels_per_coarse_pixel + self._fine_pixels_per_coarse_pixel
        col_offset = col_idx * self._fine_pixels_per_coarse_pixel + self._fine_pixels_per_coarse_pixel
        self._expand_tiepoint_array_1km_with_repeat(tiepoint_value, expanded_arr, row_offset, col_offset)

    cdef void _expand_tiepoint_array_1km_with_repeat(
            self,
            floating tiepoint_value,
            floating[:, ::1] expanded_arr,
            Py_ssize_t row_offset,
            Py_ssize_t col_offset,
    ) noexcept nogil:
        cdef Py_ssize_t length_repeat_cycle, width_repeat_cycle
        cdef Py_ssize_t row_repeat_offset, col_repeat_offset
        for length_repeat_cycle in range(self._fine_pixels_per_coarse_pixel):
            row_repeat_offset = row_offset + length_repeat_cycle
            for width_repeat_cycle in range(self._fine_pixels_per_coarse_pixel):
                col_repeat_offset = col_offset + width_repeat_cycle
                expanded_arr[row_repeat_offset, col_repeat_offset] = tiepoint_value

    cdef void _expand_tiepoint_array_5km(self, floating[:, :] input_arr, floating[:, ::1] expanded_arr) noexcept nogil:
        self._expand_tiepoint_array_5km_main(input_arr, expanded_arr)
        self._expand_tiepoint_array_5km_left(input_arr, expanded_arr)
        if self._coarse_scan_width == 270:
            self._expand_tiepoint_array_5km_270_extra_column(input_arr, expanded_arr)
        self._expand_tiepoint_array_5km_right(input_arr, expanded_arr)

    cdef void _expand_tiepoint_array_5km_main(self, floating[:, :] input_arr, floating[:, ::1] expanded_arr) noexcept nogil:
        cdef floating tiepoint_value
        cdef Py_ssize_t row_idx, col_idx, row_offset, col_offset
        for row_idx in range(input_arr.shape[0]):
            row_offset = row_idx * self._fine_pixels_per_coarse_pixel * 2
            for col_idx in range(input_arr.shape[1]):
                col_offset = col_idx * self._fine_pixels_per_coarse_pixel + self._factor_5km
                tiepoint_value = input_arr[row_idx, col_idx]
                self._expand_tiepoint_array_5km_with_repeat(
                    tiepoint_value,
                    expanded_arr,
                    row_offset,
                    col_offset,
                    self._fine_pixels_per_coarse_pixel)

    cdef void _expand_tiepoint_array_5km_270_extra_column(
            self,
            floating[:, :] input_arr,
            floating[:, ::1] expanded_arr
    ) noexcept nogil:
        """Copy an extra coarse pixel column between the main copied area and the right-most columns."""
        cdef floating tiepoint_value
        cdef Py_ssize_t row_idx, col_idx, row_offset, col_offset
        col_idx = input_arr.shape[1] - 1
        col_offset = col_idx * self._fine_pixels_per_coarse_pixel + self._fine_pixels_per_coarse_pixel + self._factor_5km
        for row_idx in range(input_arr.shape[0]):
            row_offset = row_idx * self._fine_pixels_per_coarse_pixel * 2
            tiepoint_value = input_arr[row_idx, col_idx]
            self._expand_tiepoint_array_5km_with_repeat(
                tiepoint_value,
                expanded_arr,
                row_offset,
                col_offset,
                self._fine_pixels_per_coarse_pixel)

    cdef void _expand_tiepoint_array_5km_left(
            self,
            floating[:, :] input_arr,
            floating[:, ::1] expanded_arr,
    ) noexcept nogil:
        self._expand_tiepoint_array_5km_edges(input_arr, expanded_arr, 0, 0)

    cdef void _expand_tiepoint_array_5km_right(
            self,
            floating[:, :] input_arr,
            floating[:, ::1] expanded_arr,
    ) noexcept nogil:
        self._expand_tiepoint_array_5km_edges(
            input_arr,
            expanded_arr,
            input_arr.shape[1] - 1,
            expanded_arr.shape[1] - self._factor_5km)

    cdef void _expand_tiepoint_array_5km_edges(
            self,
            floating[:, :] input_arr,
            floating[:, ::1] expanded_arr,
            Py_ssize_t course_col_idx,
            Py_ssize_t fine_col_idx,
    ) noexcept nogil:
        cdef floating tiepoint_value
        cdef Py_ssize_t row_idx, row_offset
        for row_idx in range(input_arr.shape[0]):
            row_offset = row_idx * self._fine_pixels_per_coarse_pixel * 2
            tiepoint_value = input_arr[row_idx, course_col_idx]
            self._expand_tiepoint_array_5km_with_repeat(
                tiepoint_value,
                expanded_arr,
                row_offset,
                fine_col_idx,
                self._factor_5km)

    cdef void _expand_tiepoint_array_5km_with_repeat(
            self,
            floating tiepoint_value,
            floating[:, ::1] expanded_arr,
            Py_ssize_t row_offset,
            Py_ssize_t col_offset,
            Py_ssize_t col_width,
    ) noexcept nogil:
        cdef Py_ssize_t length_repeat_cycle, width_repeat_cycle
        for length_repeat_cycle in range(self._fine_pixels_per_coarse_pixel * 2):
            for width_repeat_cycle in range(col_width):
                expanded_arr[row_offset + length_repeat_cycle,
                             col_offset + width_repeat_cycle] = tiepoint_value



"""Interpolation of geographical tiepoints for the VII products.

It follows the description provided in document "EPS-SG VII Level 1B Product Format Specification".
Tiepoints are typically subsampled by a factor 8 with respect to the pixels, along and across the satellite track.
Because of the bowtie effect, tiepoints at the scan edges are not continuous between neighbouring scans,
therefore each scan has its own edge tiepoints in the along-track direction.
However, for computational efficiency, the edge tie points that lie outside the original data point raster
are excluded from the interpolation grid which is then carried out per granule, rather than per scan
at a (very) small geolocation accuracy cost at the swath edge (investigation to quantify this ongoing).
The interpolation functions are implemented for xarray.DataArrays as input.
This version works with vii test data V2 to be released Jan 2022 which has the data stored
in alt, act (row,col) format instead of act,alt (col,row)
"""

import xarray as xr
import dask.array as da
import numpy as np

# MEAN EARTH RADIUS AS DEFINED BY IUGG
MEAN_EARTH_RADIUS = 6371008.7714  # [m]


def tie_points_interpolation(data_on_tie_points, scan_alt_tie_points, tie_points_factor):
    """Interpolate the data from the tie points to the pixel points.
    The data are provided as a list of xarray DataArray objects, allowing to interpolate on several arrays
    at the same time; however the individual arrays must have exactly the same dimensions.
    Args:
        data_on_tie_points: list of xarray DataArray objects containing the values defined on the tie points.
        scan_alt_tie_points: number of tie points along the satellite track for each scan
        tie_points_factor: sub-sampling factor of tie points wrt pixel points
    Returns:
        list of xarray DataArray objects containing the interpolated values on the pixel points.
    """
    # Extract the dimensions of the tie points array across and along track
    n_tie_alt, n_tie_act = data_on_tie_points[0].shape
    dim_alt, dim_act = data_on_tie_points[0].dims

    # Check that the number of tie points along track is multiple of the number of tie points per scan
    if n_tie_alt % scan_alt_tie_points != 0:
        raise ValueError("The number of tie points in the along-route dimension must be a multiple of %d",
                         scan_alt_tie_points)

    # Compute the number of scans
    n_scans = n_tie_alt // scan_alt_tie_points

    # Compute the dimensions of the pixel points array across and along track
    n_pixel_act = (n_tie_act - 1) * tie_points_factor
    n_pixel_alt = (n_tie_alt - 1) * tie_points_factor

    # Create the grids used for interpolation across the track
    tie_grid_act = np.arange(0, n_pixel_act + 1, tie_points_factor)
    pixel_grid_act = np.arange(0, n_pixel_act)

    # Create the grids used for the interpolation along the track (must not include the spurious points between scans)
    tie_grid_alt = np.arange(0, n_pixel_alt + 1, tie_points_factor)
    n_pixel_alt_per_scan = (scan_alt_tie_points - 1) * tie_points_factor
    pixel_grid_alt = []

    for j_scan in range(n_scans):
        start_index_scan = j_scan * scan_alt_tie_points * tie_points_factor
        pixel_grid_alt.append(np.arange(start_index_scan, start_index_scan + n_pixel_alt_per_scan))
    pixel_grid_alt = np.concatenate(pixel_grid_alt)

    # Loop on all arrays
    data_on_pixel_points = []
    for data in data_on_tie_points:

        if data.shape != (n_tie_alt, n_tie_act) or data.dims != (dim_alt, dim_act):
            raise ValueError("The dimensions of the arrays are not consistent")

        # Interpolate using the xarray interp function twice: first across, then along the scan
        # (much faster than interpolating directly in the two dimensions)
        data = data.assign_coords({dim_alt: tie_grid_alt, dim_act: tie_grid_act})
        data_pixel = data.interp({dim_alt: pixel_grid_alt}, assume_sorted=True) \
                         .interp({dim_act: pixel_grid_act}, assume_sorted=True).drop_vars([dim_alt, dim_act])

        data_on_pixel_points.append(data_pixel)

    return data_on_pixel_points



def tie_points_geo_interpolation(longitude, latitude,
                                 scan_alt_tie_points, tie_points_factor,
                                 lat_threshold_use_cartesian=60.,
                                 z_threshold_use_xy=0.8):
    """Interpolate the geographical position from the tie points to the pixel points.

    The longitude and latitude values are provided as xarray DataArray objects.

    Args:
        data_on_tie_points: list of xarray DataArray objects containing the values defined on the tie points.
        scan_alt_tie_points: number of tie points along the satellite track for each scan
        tie_points_factor: sub-sampling factor of tie points wrt pixel points

    Returns:
        list of xarray DataArray objects containing the interpolated values on the pixel points.

    Args:
        longitude: xarray DataArray containing the longitude values defined on the tie points (degrees).
        latitude: xarray DataArray containing the latitude values defined on the tie points (degrees).
        scan_alt_tie_points: number of tie points along the satellite track for each scan.
        tie_points_factor: sub-sampling factor of tie points wrt pixel points.
        lat_threshold_use_cartesian: latitude threshold to use cartesian coordinates.
        z_threshold_use_xy: z threshold to compute latitude from x and y in cartesian coordinates.

    Returns:
        two xarray DataArray objects containing the interpolated longitude and latitude values on the pixel points.

    """
    # Check that the two arrays have the same dimensions
    if longitude.shape != latitude.shape:
        raise ValueError("The dimensions of longitude and latitude don't match")

    # Determine if the interpolation should be done in cartesian or geodetic coordinates
    to_cart = np.max(np.fabs(latitude)) > lat_threshold_use_cartesian or (np.max(longitude) - np.min(longitude)) > 180.

    if to_cart:

        x, y, z = _lonlat2xyz(longitude, latitude)

        interp_x, interp_y, interp_z = tie_points_interpolation([x, y, z],
                                                                scan_alt_tie_points,
                                                                tie_points_factor)

        interp_longitude, interp_latitude = _xyz2lonlat(interp_x, interp_y, interp_z, z_threshold_use_xy)

    else:

        interp_longitude, interp_latitude = tie_points_interpolation([longitude, latitude],
                                                                     scan_alt_tie_points,
                                                                     tie_points_factor)

    return interp_longitude, interp_latitude


def _lonlat2xyz(lons, lats):
    """Convert longitudes and latitudes to cartesian coordinates.

    Args:
        lons: array containing the longitude values in degrees.
        lats: array containing the latitude values in degrees.

    Returns:
        tuple of arrays containing the x, y, and z values in meters.

    """
    lons_rad = np.deg2rad(lons)
    lats_rad = np.deg2rad(lats)
    x_coords = MEAN_EARTH_RADIUS * np.cos(lats_rad) * np.cos(lons_rad)
    y_coords = MEAN_EARTH_RADIUS * np.cos(lats_rad) * np.sin(lons_rad)
    z_coords = MEAN_EARTH_RADIUS * np.sin(lats_rad)
    return x_coords, y_coords, z_coords


def _xyz2lonlat(x_coords, y_coords, z_coords, z_threshold_use_xy=0.8):
    """Get longitudes and latitudes from cartesian coordinates.

    Args:
        x_coords: array containing the x values in meters.
        y_coords: array containing the y values in meters.
        z_coords: array containing the z values in meters.
        z_threshold_use_xy: z threshold to compute latitude from x and y in cartesian coordinates.

    Returns:
        tuple of arrays containing the longitude and latitude values in degrees.

    """
    r = np.sqrt(x_coords ** 2 + y_coords ** 2)
    thr_z = z_threshold_use_xy * MEAN_EARTH_RADIUS
    lons = np.rad2deg(np.arccos(x_coords / r)) * np.sign(y_coords)
    # Compute latitude from z at low z and from x and y at high z
    lats = xr.where(
        np.logical_and(np.less(z_coords, thr_z), np.greater(z_coords, -thr_z)),
        90. - np.rad2deg(np.arccos(z_coords / MEAN_EARTH_RADIUS)),
        np.sign(z_coords) * (90. - np.rad2deg(np.arcsin(r / MEAN_EARTH_RADIUS)))
    )
    return lons, lats



"""Interpolation of MODIS data using satellite zenith angle.

Interpolation of geographical tiepoints using the second order interpolation
scheme implemented in the CVIIRS software, as described here:
Compact VIIRS SDR Product Format User Guide (V1J)
https://www.eumetsat.int/media/45988
and
Anders Meier Soerensen, Stephan Zinke,
A tie-point zone group compaction schema for the geolocation data of S-NPP and NOAA-20 VIIRS SDRs to reduce file sizes
in memory-sensitive environments,
Applied Computing and Geosciences, Volume 6, 2020, 100025, ISSN 2590-1974,
https://doi.org/10.1016/j.acags.2020.100025.
(https://www.sciencedirect.com/science/article/pii/S2590197420300070)
"""

import warnings

from ._modis_interpolator import interpolate


def modis_1km_to_250m(lon1, lat1, satz1):
    """Interpolate MODIS geolocation from 1km to 250m resolution."""
    return interpolate(lon1, lat1, satz1,
                       coarse_resolution=1000,
                       fine_resolution=250)


def modis_1km_to_500m(lon1, lat1, satz1):
    """Interpolate MODIS geolocation from 1km to 500m resolution."""
    return interpolate(lon1, lat1, satz1,
                       coarse_resolution=1000,
                       fine_resolution=500)


def modis_5km_to_1km(lon1, lat1, satz1):
    """Interpolate MODIS geolocation from 5km to 1km resolution."""
    return interpolate(lon1, lat1, satz1,
                       coarse_resolution=5000,
                       fine_resolution=1000,
                       coarse_scan_width=lon1.shape[1])


def modis_5km_to_500m(lon1, lat1, satz1):
    """Interpolate MODIS geolocation from 5km to 500m resolution."""
    warnings.warn(
        "Interpolating 5km geolocation to 500m resolution " "may result in poor quality"
    )
    return interpolate(lon1, lat1, satz1,
                       coarse_resolution=5000,
                       fine_resolution=500,
                       coarse_scan_width=lon1.shape[1])


def modis_5km_to_250m(lon1, lat1, satz1):
    """Interpolate MODIS geolocation from 5km to 250m resolution."""
    warnings.warn(
        "Interpolating 5km geolocation to 250m resolution " "may result in poor quality"
    )
    return interpolate(lon1, lat1, satz1,
                       coarse_resolution=5000,
                       fine_resolution=250,
                       coarse_scan_width=lon1.shape[1])



"""Unit tests for python-geotiepoints: MODIS examples."""

import numpy as np
import h5py
import os

import pytest

FILENAME_250M_RESULT = os.path.join(
    os.path.dirname(__file__), '../../testdata/250m_lonlat_section_result.h5')
FILENAME_250M_INPUT = os.path.join(
    os.path.dirname(__file__), '../../testdata/250m_lonlat_section_input.h5')

FILENAME_FULL = os.path.join(
    os.path.dirname(__file__), '../../testdata/test_5_to_1_geoloc_full.h5')
FILENAME_5KM = os.path.join(
    os.path.dirname(__file__), '../../testdata/test_5_to_1_geoloc_5km.h5')

from geotiepoints import (modis5kmto1km, modis1kmto250m)

from geotiepoints import get_scene_splits


class TestUtils:
    """Class for unit testing the ancillary interpolation functions."""

    def test_get_numof_subscene_lines(self):
        """Test getting the number of sub-scene lines.

        Function is dependent on the number of CPUs and for various number of
        lines in a scan.
        """
        ncpus = 3
        scene_splits = get_scene_splits(1060, 10, ncpus)
        assert list(scene_splits) == [350, 700, 1050]


class TestMODIS:
    """Class for system testing the MODIS interpolation."""

    def test_5_to_1(self):
        """Test the 5km to 1km interpolation facility."""

        with h5py.File(FILENAME_FULL) as h5f:
            glons = h5f['longitude'][:] / 1000.
            glats = h5f['latitude'][:] / 1000.

        with h5py.File(FILENAME_5KM) as h5f:
            lons = h5f['longitude'][:] / 1000.
            lats = h5f['latitude'][:] / 1000.

        tlons, tlats = modis5kmto1km(lons, lats)

        np.testing.assert_allclose(tlons, glons, atol=0.05)
        np.testing.assert_allclose(tlats, glats, atol=0.05)

    @pytest.mark.parametrize("ncores", [None, 4])
    def test_1000m_to_250m(self, ncores):
        """Test the 1 km to 250 meter interpolation facility."""
        if ncores:
            import multiprocessing as mp
            mp.set_start_method("spawn", force=True)

        with h5py.File(FILENAME_250M_RESULT) as h5f:
            glons = h5f['longitude'][:] / 1000.
            glats = h5f['latitude'][:] / 1000.

        with h5py.File(FILENAME_250M_INPUT) as h5f:
            lons = h5f['longitude'][:] / 1000.
            lats = h5f['latitude'][:] / 1000.

        kwargs = {"cores": ncores} if ncores is not None else {}
        tlons, tlats = modis1kmto250m(lons, lats, **kwargs)
        np.testing.assert_allclose(tlons, glons, atol=0.05)
        np.testing.assert_allclose(tlats, glats, atol=0.05)



"""Tests for MODIS interpolators."""
import warnings

import numpy as np
from pyproj import Geod
import h5py
import os
import dask
import dask.array as da
import xarray as xr
import pytest
from .utils import CustomScheduler
from geotiepoints.modisinterpolator import (modis_1km_to_250m,
                                            modis_1km_to_500m,
                                            modis_5km_to_1km,
                                            modis_5km_to_500m,
                                            modis_5km_to_250m)
FILENAME_DATA = os.path.join(
    os.path.dirname(__file__), '../../testdata/modis_test_data.h5')


def _to_dask(arr):
    return da.from_array(arr, chunks=4096)


def _to_da(arr):
    return xr.DataArray(_to_dask(arr), dims=['y', 'x'])


def _load_h5_geo_vars(*var_names):
    h5f = h5py.File(FILENAME_DATA, 'r')
    return tuple(h5f[var_name] for var_name in var_names)


def load_1km_lonlat_as_numpy():
    lon1, lat1 = _load_h5_geo_vars('lon_1km', 'lat_1km')
    return lon1[:], lat1[:]


def load_1km_lonlat_as_dask():
    lon1, lat1 = _load_h5_geo_vars('lon_1km', 'lat_1km')
    return _to_dask(lon1), _to_dask(lat1)


def load_1km_lonlat_as_xarray_dask():
    lon1, lat1 = _load_h5_geo_vars('lon_1km', 'lat_1km')
    return _to_da(lon1), _to_da(lat1)


def load_1km_lonlat_satz_as_xarray_dask():
    lon1, lat1, satz1 = _load_h5_geo_vars('lon_1km', 'lat_1km', 'satz_1km')
    return _to_da(lon1), _to_da(lat1), _to_da(satz1)


def load_5km_lonlat_satz1_as_xarray_dask():
    lon1, lat1, satz1 = _load_h5_geo_vars('lon_1km', 'lat_1km', 'satz_1km')
    lon5 = lon1[2::5, 2::5]
    lat5 = lat1[2::5, 2::5]
    satz5 = satz1[2::5, 2::5]
    return _to_da(lon5), _to_da(lat5), _to_da(satz5)


def load_l2_5km_lonlat_satz1_as_xarray_dask():
    lon1, lat1, satz1 = _load_h5_geo_vars('lon_1km', 'lat_1km', 'satz_1km')
    lon5 = lon1[2::5, 2:-5:5]
    lat5 = lat1[2::5, 2:-5:5]
    satz5 = satz1[2::5, 2:-5:5]
    return _to_da(lon5), _to_da(lat5), _to_da(satz5)


def load_500m_lonlat_expected_as_xarray_dask():
    h5f = h5py.File(FILENAME_DATA, 'r')
    lon500 = _to_da(h5f['lon_500m'])
    lat500 = _to_da(h5f['lat_500m'])
    return lon500, lat500


def load_250m_lonlat_expected_as_xarray_dask():
    h5f = h5py.File(FILENAME_DATA, 'r')
    lon250 = _to_da(h5f['lon_250m'])
    lat250 = _to_da(h5f['lat_250m'])
    return lon250, lat250


def assert_geodetic_distance(
        lons_actual: np.ndarray,
        lats_actual: np.ndarray,
        lons_desired: np.ndarray,
        lats_desired: np.ndarray,
        max_distance_diff: float,
) -> None:
    """Check that the geodetic distance between two sets of coordinates is smaller than a threshold.

    Args:
        lons_actual: Longitude array produced by interpolation being tested.
        lats_actual: Latitude array produced by interpolation being tested.
        lons_desired: Longitude array of expected/truth coordinates.
        lats_desired: Latitude array of expected/truth coordinates.
        max_distance_diff: Limit of allowed distance difference in meters.

    """
    g = Geod(ellps="WGS84")
    _, _, dist = g.inv(lons_actual, lats_actual, lons_desired, lats_desired)
    np.testing.assert_array_less(
        dist, max_distance_diff,
        err_msg=f"Coordinates are greater than {max_distance_diff} geodetic "
                "meters from the expected coordinates.")


@pytest.mark.parametrize(
    ("input_func", "exp_func", "interp_func", "dist_max", "exp_5km_warning"),
    [
        (load_1km_lonlat_satz_as_xarray_dask, load_500m_lonlat_expected_as_xarray_dask, modis_1km_to_500m, 5, False),
        (load_1km_lonlat_satz_as_xarray_dask, load_250m_lonlat_expected_as_xarray_dask, modis_1km_to_250m, 8.30, False),
        (load_5km_lonlat_satz1_as_xarray_dask, load_1km_lonlat_as_xarray_dask, modis_5km_to_1km, 25, False),
        (load_l2_5km_lonlat_satz1_as_xarray_dask, load_1km_lonlat_as_xarray_dask, modis_5km_to_1km, 110, False),
        (load_5km_lonlat_satz1_as_xarray_dask, load_500m_lonlat_expected_as_xarray_dask, modis_5km_to_500m,
         19500, True),
        (load_5km_lonlat_satz1_as_xarray_dask, load_250m_lonlat_expected_as_xarray_dask, modis_5km_to_250m,
         25800, True),
    ]
)
def test_sat_angle_based_interp(input_func, exp_func, interp_func, dist_max, exp_5km_warning):
    lon1, lat1, satz1 = input_func()
    lons_exp, lats_exp = exp_func()

    # when working with dask arrays, we shouldn't compute anything
    with dask.config.set(scheduler=CustomScheduler(0)), warnings.catch_warnings(record=True) as warns:
        lons, lats = interp_func(lon1, lat1, satz1)
    has_5km_warning = any("may result in poor quality" in str(w.message) for w in warns)
    if exp_5km_warning:
        assert has_5km_warning
    else:
        assert not has_5km_warning

    if hasattr(lons, "compute"):
        lons, lats = da.compute(lons, lats)
    assert_geodetic_distance(lons, lats, lons_exp, lats_exp, dist_max)
    assert not np.any(np.isnan(lons))
    assert not np.any(np.isnan(lats))


def test_sat_angle_based_interp_nan_handling():
    # See GH #19
    lon1, lat1, satz1 = load_1km_lonlat_satz_as_xarray_dask()
    satz1 = _to_da(abs(np.linspace(-65.4, 65.4, 1354, dtype=np.float32)).repeat(20).reshape(-1, 20).T)
    lons, lats = modis_1km_to_500m(lon1, lat1, satz1)
    assert not np.any(np.isnan(lons.compute()))
    assert not np.any(np.isnan(lats.compute()))


def test_poles_datum():
    orig_lon, lat1, satz1 = load_1km_lonlat_satz_as_xarray_dask()
    lon1 = orig_lon + 180
    lon1 = xr.where(lon1 > 180, lon1 - 360, lon1)

    lat5 = lat1[2::5, 2::5]
    lon5 = lon1[2::5, 2::5]
    satz5 = satz1[2::5, 2::5]
    lons, lats = modis_5km_to_1km(lon5, lat5, satz5)

    lons = lons + 180
    lons = xr.where(lons > 180, lons - 360, lons)
    assert_geodetic_distance(lons, lats, orig_lon, lat1, 25.0)



"""Setting up the geo_interpolator project."""

import sys

from setuptools import setup, find_packages
import versioneer
import numpy as np
from Cython.Build import build_ext
from Cython.Distutils import Extension

requirements = ['numpy', 'scipy', 'pandas']
test_requires = ['pytest', 'pytest-cov', 'h5py', 'xarray', 'dask[array]', 'pyproj', "pyresample"]

if sys.platform.startswith("win"):
    extra_compile_args = []
else:
    extra_compile_args = ["-O3"]

EXTENSIONS = [
    Extension(
        'geotiepoints.multilinear_cython',
        sources=['geotiepoints/multilinear_cython.pyx'],
        extra_compile_args=extra_compile_args,
        include_dirs=[np.get_include()],
    ),
    Extension(
        'geotiepoints._modis_interpolator',
        sources=['geotiepoints/_modis_interpolator.pyx'],
        extra_compile_args=extra_compile_args,
        include_dirs=[np.get_include()],
    ),
    Extension(
        'geotiepoints._simple_modis_interpolator',
        sources=['geotiepoints/_simple_modis_interpolator.pyx'],
        extra_compile_args=extra_compile_args,
        include_dirs=[np.get_include()],
    ),
    Extension(
        'geotiepoints._modis_utils',
        sources=['geotiepoints/_modis_utils.pyx'],
        extra_compile_args=extra_compile_args,
        include_dirs=[np.get_include()],
    ),
]


try:
    sys.argv.remove("--cython-coverage")
    cython_coverage = True
except ValueError:
    cython_coverage = False


cython_directives = {
    "language_level": "3",
    "freethreading_compatible": True,
}
define_macros = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
if cython_coverage:
    print("Enabling directives/macros for Cython coverage support")
    cython_directives.update({
        "linetrace": True,
        "profile": True,
    })
    define_macros.extend([
        ("CYTHON_TRACE", "1"),
        ("CYTHON_TRACE_NOGIL", "1"),
    ])
for ext in EXTENSIONS:
    ext.define_macros = define_macros
    ext.cython_directives.update(cython_directives)

cmdclass = versioneer.get_cmdclass(cmdclass={"build_ext": build_ext})

with open('README.md', 'r') as readme:
    README = readme.read()

if __name__ == "__main__":
    setup(name='python-geotiepoints',
          version=versioneer.get_version(),
          description='Interpolation of geographic tiepoints in Python',
          long_description=README,
          long_description_content_type='text/markdown',
          author='Adam Dybbroe, Martin Raspaud',
          author_email='martin.raspaud@smhi.se',
          classifiers=[
              "Development Status :: 5 - Production/Stable",
              "Intended Audience :: Science/Research",
              "Operating System :: OS Independent",
              "Programming Language :: Python",
              "Programming Language :: Cython",
              "Topic :: Scientific/Engineering",
              "Programming Language :: Python :: Free Threading :: 1 - Unstable",
          ],
          license="Apache-2.0",
          license_files=["LICENSE.txt"],
          url="https://github.com/pytroll/python-geotiepoints",
          packages=find_packages(),
          python_requires='>=3.11',
          cmdclass=cmdclass,
          install_requires=requirements,
          ext_modules=EXTENSIONS,
          extras_require={
              "tests": test_requires,
          },
          zip_safe=False
          )



"""Generic interpolation routines."""

from abc import ABC, abstractmethod
from functools import partial
import numpy as np
from scipy.interpolate import RectBivariateSpline, splev, splrep, RegularGridInterpolator


def generic_modis5kmto1km(*data5km):
    """Get 1km data for modis from 5km tiepoints."""
    cols5km = np.arange(2, 1354, 5)
    cols1km = np.arange(1354)
    lines = data5km[0].shape[0] * 5
    rows5km = np.arange(2, lines, 5)
    rows1km = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3

    satint = Interpolator(list(data5km),
                          (rows5km, cols5km),
                          (rows1km, cols1km),
                          along_track_order,
                          cross_track_order,
                          chunk_size=10)
    satint.fill_borders("y", "x")
    return satint.interpolate()

# NOTE: extrapolate on a sphere ?


def _linear_extrapolate(pos, data, xev):
    """Perform linear extrapolation.

    >>> import numpy as np
    >>> pos = np.array([1, 2])
    >>> data = np.arange(10).reshape((2, 5), order="F")
    >>> xev = 5
    >>> retv = _linear_extrapolate(pos, data, xev)
    >>> print([val for val in retv])
    [4.0, 6.0, 8.0, 10.0, 12.0]
    >>> xev = 0
    >>> retv = _linear_extrapolate(pos, data, xev)
    >>> print([val for val in retv])
    [-1.0, 1.0, 3.0, 5.0, 7.0]
    """
    if len(data) != 2 or len(pos) != 2:
        raise ValueError("len(pos) and the number of lines of data"
                         " must be 2.")

    return data[1] + ((xev - pos[1]) / (1.0 * (pos[0] - pos[1])) * (data[0] - data[1]))


class Interpolator:
    """Handles interpolation of data from a grid of tie points.

    It is preferable to have tie-points out till the edges if the tiepoint grid,
    but a method is provided to extrapolate linearly the tiepoints to the borders of the
    grid. The extrapolation is done automatically if it seems necessary.

    Uses numpy and scipy.

    The constructor takes in the tiepointed data as *data*, the *tiepoint_grid* and the desired *final_grid*. As
    optional arguments, one can provide *kx_* and *ky_* as interpolation orders (in x and y directions respectively),
    and the *chunksize* if the data has to be handled by pieces along the y axis (this affects how the extrapolator
    behaves). If *chunksize* is set, don't forget to adjust the interpolation orders accordingly: the interpolation
    is indeed done globaly (not chunkwise).
    """

    def __init__(self, data, tiepoint_grid, final_grid,
                 kx_=1, ky_=1, chunk_size=0):
        self.row_indices = tiepoint_grid[0]
        self.col_indices = tiepoint_grid[1]
        self.hrow_indices = final_grid[0]
        self.hcol_indices = final_grid[1]
        self.chunk_size = chunk_size
        if not isinstance(data, (tuple, list)):
            self.tie_data = [data]
        else:
            self.tie_data = list(data)

        self.new_data = [[] for _ in self.tie_data]
        self.kx_, self.ky_ = kx_, ky_

    def fill_borders(self, *args):
        """Extrapolate tiepoint lons and lats to fill in the border of the chunks."""
        to_run = []
        cases = {"y": self._fill_row_borders,
                 "x": self._fill_col_borders}
        for dim in args:
            try:
                to_run.append(cases[dim])
            except KeyError:
                raise NameError("Unrecognized dimension: " + str(dim))

        for fun in to_run:
            fun()

    def _extrapolate_cols(self, data, first=True, last=True):
        """Extrapolate the column of data, to get the first and last together with the data."""
        if first:
            pos = self.col_indices[:2]
            first_column = _linear_extrapolate(pos,
                                               (data[:, 0], data[:, 1]),
                                               self.hcol_indices[0])
        if last:
            pos = self.col_indices[-2:]
            last_column = _linear_extrapolate(pos,
                                              (data[:, -2], data[:, -1]),
                                              self.hcol_indices[-1])

        if first and last:
            return np.hstack((np.expand_dims(first_column, 1),
                              data,
                              np.expand_dims(last_column, 1)))
        elif first:
            return np.hstack((np.expand_dims(first_column, 1),
                              data))
        elif last:
            return np.hstack((data,
                              np.expand_dims(last_column, 1)))
        else:
            return data

    def _fill_col_borders(self):
        """Add the first and last column to the data by extrapolation."""
        first = True
        last = True
        if self.col_indices[0] == self.hcol_indices[0]:
            first = False
        if self.col_indices[-1] == self.hcol_indices[-1]:
            last = False
        for num, data in enumerate(self.tie_data):
            self.tie_data[num] = self._extrapolate_cols(data, first, last)

        if first and last:
            self.col_indices = np.concatenate((np.array([self.hcol_indices[0]]),
                                               self.col_indices,
                                               np.array([self.hcol_indices[-1]])))
        elif first:
            self.col_indices = np.concatenate((np.array([self.hcol_indices[0]]),
                                               self.col_indices))
        elif last:
            self.col_indices = np.concatenate((self.col_indices,
                                               np.array([self.hcol_indices[-1]])))

    def _extrapolate_rows(self, data, row_indices, first_index, last_index):
        """Extrapolate the rows of data, to get the first and last together with the data."""
        pos = row_indices[:2]
        first_row = _linear_extrapolate(pos,
                                        (data[0, :], data[1, :]),
                                        first_index)
        pos = row_indices[-2:]
        last_row = _linear_extrapolate(pos,
                                       (data[-2, :], data[-1, :]),
                                       last_index)
        return np.vstack((np.expand_dims(first_row, 0),
                          data,
                          np.expand_dims(last_row, 0)))

    def _fill_row_borders(self):
        """Add the first and last rows to the data by extrapolation."""
        lines = len(self.hrow_indices)
        chunk_size = self.chunk_size or lines
        factor = len(self.hrow_indices) / len(self.row_indices)

        tmp_data = []
        for _num in range(len(self.tie_data)):
            tmp_data.append([])
        row_indices = []

        for index in range(0, lines, chunk_size):
            indices = np.logical_and(self.row_indices >= index / factor,
                                     self.row_indices < (index + chunk_size) / factor)
            ties = np.argwhere(indices).squeeze()
            tiepos = self.row_indices[indices].squeeze()

            for num, data in enumerate(self.tie_data):
                to_extrapolate = data[ties, :]
                if len(to_extrapolate) > 0:
                    extrapolated = self._extrapolate_rows(to_extrapolate,
                                                          tiepos,
                                                          self.hrow_indices[
                                                              index],
                                                          self.hrow_indices[index + chunk_size - 1])
                    tmp_data[num].append(extrapolated)

            row_indices.append(np.array([self.hrow_indices[index]]))
            row_indices.append(tiepos)
            row_indices.append(np.array([self.hrow_indices[index + chunk_size - 1]]))

        for num in range(len(self.tie_data)):
            self.tie_data[num] = np.vstack(tmp_data[num])
        self.row_indices = np.concatenate(row_indices)

    def _interp(self):
        """Interpolate the cartesian coordinates."""
        if np.array_equal(self.hrow_indices, self.row_indices):
            return self._interp1d()

        for num, data in enumerate(self.tie_data):
            spl = RectBivariateSpline(self.row_indices,
                                      self.col_indices,
                                      data,
                                      s=0,
                                      kx=self.kx_,
                                      ky=self.ky_)

            self.new_data[num] = spl(self.hrow_indices, self.hcol_indices, grid=True)

    def _interp1d(self):
        """Interpolate in one dimension."""
        lines = len(self.hrow_indices)

        for num, data in enumerate(self.tie_data):
            self.new_data[num] = np.empty((len(self.hrow_indices),
                                           len(self.hcol_indices)),
                                          data.dtype)

            for cnt in range(lines):
                tck = splrep(self.col_indices, data[cnt, :], k=self.ky_, s=0)
                self.new_data[num][cnt, :] = splev(
                    self.hcol_indices, tck, der=0)

    def interpolate(self):
        """Do the interpolation, and return resulting longitudes and latitudes."""
        self._interp()

        return self.new_data


class AbstractSingleInterpolator(ABC):
    """An abstract interpolator for a single 2d data array."""

    def __init__(self, points, values, scipy_interpolator, **interpolator_init_kwargs):
        """Set up the interpolator.

        *kwargs* are passed to the underlying scipy interpolator instance.
        So for example, to allow extrapolation, the kwargs can be `bounds_error=False, fill_value=None`.
        """
        self.interpolator = scipy_interpolator(points, values, **interpolator_init_kwargs)
        self.points = points
        self.values = values

    def interpolate(self, fine_points, chunks=None, **interpolator_call_kwargs):
        """Interpolate the value points to the *fine_points* grid.

        Args:
            fine_points: the points on the target grid to use, as one dimensional vectors for each dimension.
            chunks: If not None, a lazy (dask-based) interpolation will be performed using the chunk sizes specified.
                    The result will be a dask array in this case. Defaults to None.
            interpolator_kwargs: The keyword arguments to pass to the underlying scipy interpolator.
        """
        if chunks is not None:
            res = self.interpolate_dask(fine_points, chunks=chunks, **interpolator_call_kwargs)
        else:
            res = self.interpolate_numpy(fine_points, **interpolator_call_kwargs)

        return res

    def interpolate_dask(self, fine_points, chunks, **interpolator_call_kwargs):
        """Interpolate (lazily) to a dask array."""
        from dask.base import tokenize
        import dask.array as da
        from dask.array.core import normalize_chunks
        v_fine_points, h_fine_points = fine_points
        shape = len(v_fine_points), len(h_fine_points)

        chunks = normalize_chunks(chunks, shape, dtype=self.values.dtype)

        token = tokenize(chunks, self.points, self.values, fine_points, interpolator_call_kwargs)
        name = 'interpolate-' + token

        interpolate_slices = partial(self.interpolate_slices, **interpolator_call_kwargs)

        dskx = {(name, ) + position: (interpolate_slices,
                                      slices)
                for position, slices in _enumerate_chunk_slices(chunks)}

        res = da.Array(dskx, name, shape=list(shape),
                       chunks=chunks,
                       dtype=self.values.dtype)
        return res

    @abstractmethod
    def interpolate_numpy(self, fine_points, **interpolator_call_kwargs):
        """Interpolate to a numpy array."""
        raise NotImplementedError

    def interpolate_slices(self, fine_points, **interpolator_call_kwargs):
        """Interpolate using slices.

        *fine_points* are a tuple of slices for the y and x dimensions
        """
        slice_y, slice_x = fine_points
        points_y = np.arange(slice_y.start, slice_y.stop)
        points_x = np.arange(slice_x.start, slice_x.stop)
        fine_points = points_y, points_x

        return self.interpolate_numpy(fine_points, **interpolator_call_kwargs)


def _enumerate_chunk_slices(chunks):
    """Enumerate chunks with slices."""
    for position in np.ndindex(tuple(map(len, (chunks)))):
        slices = []
        for pos, chunk in zip(position, chunks):
            chunk_size = chunk[pos]
            offset = sum(chunk[:pos])
            slices.append(slice(offset, offset + chunk_size))

        yield (position, slices)


class SingleGridInterpolator(AbstractSingleInterpolator):
    """A regular grid interpolator for a single 2d data array."""

    def __init__(self, *args, **interpolator_init_kwargs):
        """Set up the grid interpolator."""
        super().__init__(*args, scipy_interpolator=RegularGridInterpolator, **interpolator_init_kwargs)

    def interpolate_numpy(self, fine_points, **interpolator_call_kwargs):
        """Interpolate to a numpy array."""
        fine_x, fine_y = np.meshgrid(*fine_points, indexing='ij')
        return self.interpolator((fine_x, fine_y), **interpolator_call_kwargs).astype(self.values.dtype)


class SingleSplineInterpolator(AbstractSingleInterpolator):
    """An spline interpolator for a single 2d data array."""

    def __init__(self, points, values, **interpolator_init_kwargs):
        """Set up the spline interpolator."""
        self.interpolator = RectBivariateSpline(*points, values, **interpolator_init_kwargs)
        self.points = points
        self.values = values

    def interpolate_numpy(self, fine_points, **interpolator_call_kwargs):
        """Interpolate to a numpy array."""
        return self.interpolator(*fine_points, **interpolator_call_kwargs).astype(self.values.dtype)


class AbstractMultipleInterpolator(ABC):  # noqa: B024
    """Abstract interpolator that works on mulitple arrays."""

    def __init__(self, interpolator, tie_points, *data, **interpolator_init_kwargs):
        """Set up the interpolator from the multiple `data` arrays."""
        self.interpolators = []
        for values in data:
            self.interpolators.append(interpolator(tie_points, values, **interpolator_init_kwargs))

    def interpolate(self, fine_points, **kwargs):
        """Interpolate the data.

        The keyword arguments will be passed on to SingleGridInterpolator's interpolate function.
        """
        return (interpolator.interpolate(fine_points, **kwargs) for interpolator in self.interpolators)

    def interpolate_to_shape(self, shape, **interpolator_call_kwargs):
        """Interpolate to a given *shape*."""
        fine_points = [np.arange(size) for size in shape]
        return self.interpolate(fine_points, **interpolator_call_kwargs)


class MultipleGridInterpolator(AbstractMultipleInterpolator):
    """Grid interpolator that works on multiple data arrays."""

    def __init__(self, tie_points, *data, **interpolator_init_kwargs):
        """Set up the interpolator from the multiple `data` arrays."""
        super().__init__(SingleGridInterpolator, tie_points, *data, **interpolator_init_kwargs)


class MultipleSplineInterpolator(AbstractMultipleInterpolator):
    """Spline interpolator that works on multiple data arrays."""

    def __init__(self, tie_points, *data, **interpolator_init_kwargs):
        """Set up the interpolator from the multiple `data` arrays."""
        super().__init__(SingleSplineInterpolator, tie_points, *data, **interpolator_init_kwargs)



#!/usr/bin/env python
"""Generate MODIS interpolation test data from real input data.

This script is used to generate the "testdata/modis_test_data.h5" file that is
used to validate the various modis interpolation algorithms in
python-geotiepoints. The test data file consists of 1km "truth" longitude and
latitude arrays from an input MOD03 HDF4 file and interpolated longitude and
latitude arrays at 500m and 250m resolution. The interpolation is done using
the CVIIRS based algorithm in ``geotiepoints/modisinterpolator.py``.
The CVIIRS algorithm was used as opposed to the "simple" or other interpolation
methods due to the smoother interpolation between pixels (no linear "steps").

MOD03 files geolocation data is terrain corrected. This means that the
interpolation methods currently in python-geotiepoints can't produce an
exact matching result for a round trip test of 1km (truth) ->
5km (every 5th pixel) -> 1km (interpolation result).
The input MOD03 test data was chosen due to its lack of varying terrain
(almost entirely ocean view) to minimize error/differences between the
1km truth and 1km interpolation results.

To limit size of the test data file and the reduce the execution time of tests
the test data is limited to the last 2 scans (20 rows of 1km data) of the
provided input data.

"""
import os
import sys
from datetime import datetime

import h5py
import numpy as np
from pyhdf.SD import SD, SDC
import xarray as xr
import dask.array as da

from geotiepoints.modisinterpolator import modis_1km_to_500m, modis_1km_to_250m


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", required=True,
                        help="Input MOD03 geolocation HDF4 filename to read 1km lon/lat data from.")
    parser.add_argument("-o", "--output", default="modis_test_data.h5",
                        help="Output test data HDF5 filename being created")
    args = parser.parse_args()

    num_1km_rows = 20
    lons_1km, lats_1km, satz_1km = _get_1km_lon_lat_satz_from_mod03(args.input)
    lons_1km = lons_1km[-num_1km_rows:]
    lats_1km = lats_1km[-num_1km_rows:]
    satz_1km = satz_1km[-num_1km_rows:]
    lons_1km = xr.DataArray(da.from_array(lons_1km), dims=("y", "x"))
    lats_1km = xr.DataArray(da.from_array(lats_1km), dims=("y", "x"))
    satz_1km = xr.DataArray(da.from_array(satz_1km), dims=("y", "x"))

    with h5py.File(args.output, "w") as output_h:
        lons_500m, lats_500m = modis_1km_to_500m(lons_1km, lats_1km, satz_1km)
        lons_500m = lons_500m.astype(np.float32, copy=False)
        lats_500m = lats_500m.astype(np.float32, copy=False)

        lons_250m, lats_250m = modis_1km_to_250m(lons_1km, lats_1km, satz_1km)
        lons_250m = lons_250m.astype(np.float32, copy=False)
        lats_250m = lats_250m.astype(np.float32, copy=False)

        output_h.create_dataset("lon_1km", data=lons_1km, compression="gzip", compression_opts=9, shuffle=True)
        output_h.create_dataset("lat_1km", data=lats_1km, compression="gzip", compression_opts=9, shuffle=True)
        output_h.create_dataset("satz_1km", data=satz_1km, compression="gzip", compression_opts=9, shuffle=True)
        output_h.create_dataset("lon_500m", data=lons_500m, compression="gzip", compression_opts=9, shuffle=True)
        output_h.create_dataset('lat_500m', data=lats_500m, compression="gzip", compression_opts=9, shuffle=True)
        output_h.create_dataset("lon_250m", data=lons_250m, compression="gzip", compression_opts=9, shuffle=True)
        output_h.create_dataset("lat_250m", data=lats_250m, compression="gzip", compression_opts=9, shuffle=True)
        output_h.attrs["1km_data_origin"] = os.path.basename(args.input)
        output_h.attrs["description"] = (
            "MODIS interpolation test data for the python-geotiepoints package. "
            "The 1 km data is taken directly from a MOD03 file. The 250m and "
            "500m is generated using the cviirs-based algorithm in "
            "`geotiepoints/modisinterpolator.py`. For more information see "
            "the generation script in `testdata/create_modis_test_data.py` in "
            "the python-geotiepoints git repository."
        )
        output_h.attrs["creation_date"] = datetime.utcnow().strftime("%Y-%m-%d")


def _get_1km_lon_lat_satz_from_mod03(hdf4_filename: str) -> tuple:
    h = SD(hdf4_filename, mode=SDC.READ)
    lon_var = h.select("Longitude")
    lat_var = h.select("Latitude")
    sat_zen_var = h.select("SensorZenith")

    # ensure 32-bit float
    lon_data = lon_var[:].astype(np.float32, copy=False)
    lat_data = lat_var[:].astype(np.float32, copy=False)
    sat_zen_attrs = sat_zen_var.attributes()
    scale_factor = sat_zen_attrs.get("scale_factor", 1.0)
    add_offset = sat_zen_attrs.get("add_offset", 0.0)
    sat_zen_data = (sat_zen_var[:] * scale_factor + add_offset).astype(np.float32, copy=False)

    return lon_data, lat_data, sat_zen_data


if __name__ == "__main__":
    sys.exit(main())



# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False, initializedcheck=False, nonecheck=False
cimport numpy as np

ctypedef fused floating:
    np.float32_t
    np.float64_t

cdef void lonlat2xyz(
        floating[:, ::1] lons,
        floating[:, ::1] lats,
        floating[:, :, ::1] xyz,
) noexcept nogil

cdef void xyz2lonlat(
        floating[:, :, ::1] xyz,
        floating[:, ::1] lons,
        floating[:, ::1] lats,
        bint low_lat_z=*,
        floating thr=*,
) noexcept nogil

cdef floating rad2deg(floating x) noexcept nogil
cdef floating deg2rad(floating x) noexcept nogil
